from gettext import gettext as _
import os

from rest_framework import fields, serializers
from rest_framework.validators import UniqueValidator
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from pulpcore.app import models, settings
from pulpcore.app.serializers import (
    DetailIdentityField,
    DetailRelatedField,
    LabelsField,
    LatestVersionField,
    ModelSerializer,
    RepositoryVersionIdentityField,
    RepositoryVersionRelatedField,
    RepositoryVersionsIdentityFromRepositoryField,
    ValidateFieldsMixin,
)


class RepositorySerializer(ModelSerializer):
    pulp_href = DetailIdentityField(view_name_pattern=r"repositories(-.*/.*)-detail")
    pulp_labels = LabelsField(required=False)
    versions_href = RepositoryVersionsIdentityFromRepositoryField()
    latest_version_href = LatestVersionField()
    name = serializers.CharField(
        help_text=_("A unique name for this repository."),
        validators=[UniqueValidator(queryset=models.Repository.objects.all())],
    )
    description = serializers.CharField(
        help_text=_("An optional description."), required=False, allow_null=True
    )
    retain_repo_versions = serializers.IntegerField(
        help_text=_(
            "Retain X versions of the repository. Default is null which retains all versions."
            " This is provided as a tech preview in Pulp 3 and may change in the future."
        ),
        allow_null=True,
        required=False,
        min_value=1,
    )
    remote = DetailRelatedField(
        help_text=_("An optional remote to use by default when syncing."),
        view_name_pattern=r"remotes(-.*/.*)-detail",
        queryset=models.Remote.objects.all(),
        required=False,
        allow_null=True,
    )

    def validate_remote(self, value):
        if value and type(value) not in self.Meta.model.REMOTE_TYPES:
            raise serializers.ValidationError(
                detail=_("Type for Remote '{}' does not match Repository.").format(value.name)
            )

        return value

    class Meta:
        model = models.Repository
        fields = ModelSerializer.Meta.fields + (
            "versions_href",
            "pulp_labels",
            "latest_version_href",
            "name",
            "description",
            "retain_repo_versions",
            "remote",
        )


class RemoteSerializer(ModelSerializer):
    """
    Every remote defined by a plugin should have a Remote serializer that inherits from this
    class. Please import from `pulpcore.plugin.serializers` rather than from this module directly.
    """

    pulp_href = DetailIdentityField(view_name_pattern=r"remotes(-.*/.*)-detail")
    pulp_labels = LabelsField(required=False)
    name = serializers.CharField(
        help_text=_("A unique name for this remote."),
        validators=[UniqueValidator(queryset=models.Remote.objects.all())],
    )
    url = serializers.CharField(help_text="The URL of an external content source.")
    ca_cert = serializers.CharField(
        help_text="A PEM encoded CA certificate used to validate the server "
        "certificate presented by the remote server.",
        required=False,
        allow_null=True,
    )
    client_cert = serializers.CharField(
        help_text="A PEM encoded client certificate used for authentication.",
        required=False,
        allow_null=True,
    )
    client_key = serializers.CharField(
        help_text="A PEM encoded private key used for authentication.",
        required=False,
        allow_null=True,
        write_only=True,
    )
    tls_validation = serializers.BooleanField(
        help_text="If True, TLS peer validation must be performed.", required=False
    )
    proxy_url = serializers.CharField(
        help_text="The proxy URL. Format: scheme://host:port",
        required=False,
        allow_null=True,
    )
    proxy_username = serializers.CharField(
        help_text="The username to authenticte to the proxy.",
        required=False,
        allow_null=True,
        write_only=True,
    )
    proxy_password = serializers.CharField(
        help_text="The password to authenticte to the proxy.",
        required=False,
        allow_null=True,
        write_only=True,
        style={"input_type": "password"},
    )
    username = serializers.CharField(
        help_text="The username to be used for authentication when syncing.",
        required=False,
        allow_null=True,
        write_only=True,
    )
    password = serializers.CharField(
        help_text="The password to be used for authentication when syncing.",
        required=False,
        allow_null=True,
        write_only=True,
        style={"input_type": "password"},
    )
    pulp_last_updated = serializers.DateTimeField(
        help_text="Timestamp of the most recent update of the remote.", read_only=True
    )
    download_concurrency = serializers.IntegerField(
        help_text=(
            "Total number of simultaneous connections. If not set then the default "
            "value will be used."
        ),
        allow_null=True,
        required=False,
        min_value=1,
    )
    max_retries = serializers.IntegerField(
        help_text=(
            "Maximum number of retry attempts after a download failure. If not set then the "
            "default value (3) will be used."
        ),
        required=False,
        allow_null=True,
    )
    policy = serializers.ChoiceField(
        help_text="The policy to use when downloading content.",
        choices=((models.Remote.IMMEDIATE, "When syncing, download all metadata and content now.")),
        default=models.Remote.IMMEDIATE,
    )

    total_timeout = serializers.FloatField(
        allow_null=True,
        required=False,
        help_text=(
            "aiohttp.ClientTimeout.total (q.v.) for download-connections. The default is null, "
            "which will cause the default from the aiohttp library to be used."
        ),
        min_value=0.0,
    )
    connect_timeout = serializers.FloatField(
        allow_null=True,
        required=False,
        help_text=(
            "aiohttp.ClientTimeout.connect (q.v.) for download-connections. The default is null, "
            "which will cause the default from the aiohttp library to be used."
        ),
        min_value=0.0,
    )
    sock_connect_timeout = serializers.FloatField(
        allow_null=True,
        required=False,
        help_text=(
            "aiohttp.ClientTimeout.sock_connect (q.v.) for download-connections. The default is "
            "null, which will cause the default from the aiohttp library to be used."
        ),
        min_value=0.0,
    )
    sock_read_timeout = serializers.FloatField(
        allow_null=True,
        required=False,
        help_text=(
            "aiohttp.ClientTimeout.sock_read (q.v.) for download-connections. The default is "
            "null, which will cause the default from the aiohttp library to be used."
        ),
        min_value=0.0,
    )
    headers = serializers.ListField(
        child=serializers.DictField(),
        help_text=_("Headers for aiohttp.Clientsession"),
        required=False,
    )
    rate_limit = serializers.IntegerField(
        help_text=_("Limits total download rate in requests per second"),
        allow_null=True,
        required=False,
    )

    def validate_url(self, url):
        """
        Check if the 'url' is a ``file://`` path, and if so, ensure it's an ALLOWED_IMPORT_PATH.

        The ALLOWED_IMPORT_PATH is specified as a Pulp setting.

        Args:
            url: The user-provided value for 'url' to be validated.

        Raises:
            ValidationError: When the url starts with `file://`, but is not a subfolder of a path in
                the ALLOWED_IMPORT_PATH setting.

        Returns:
            The validated value.
        """
        if not url.lower().startswith("file://"):
            return url

        user_path = url[7:]
        if not os.path.isabs(user_path):
            raise serializers.ValidationError(
                _("The path '{}' needs to be an absolute pathname.").format(user_path)
            )

        user_provided_realpath = os.path.realpath(user_path)

        for allowed_path in settings.ALLOWED_IMPORT_PATHS:
            if user_provided_realpath.startswith(allowed_path):
                return url

        raise serializers.ValidationError(
            _("The path '{}' does not start with any of the allowed import paths").format(user_path)
        )

    def validate_proxy_url(self, value):
        """
        Check, that the proxy_url does not contain credentials.
        """
        if value and "@" in value:
            raise serializers.ValidationError(_("proxy_url must not contain credentials"))
        return value

    def validate(self, data):
        """
        Check, that proxy credentials are only provided completely and if a proxy is configured.
        """
        data = super().validate(data)

        proxy_url = self.instance.proxy_url if self.partial else None
        proxy_url = data.get("proxy_url", proxy_url)
        proxy_username = self.instance.proxy_username if self.partial else None
        proxy_username = data.get("proxy_username", proxy_username)
        proxy_password = self.instance.proxy_password if self.partial else None
        proxy_password = data.get("proxy_password", proxy_password)

        if (proxy_username or proxy_password) and not proxy_url:
            raise serializers.ValidationError(
                _("proxy credentials cannot be specified without a proxy")
            )

        if bool(proxy_username) is not bool(proxy_password):
            raise serializers.ValidationError(
                _("proxy username and password can only be specified together")
            )

        return data

    class Meta:
        abstract = True
        model = models.Remote
        fields = ModelSerializer.Meta.fields + (
            "name",
            "url",
            "ca_cert",
            "client_cert",
            "client_key",
            "tls_validation",
            "proxy_url",
            "proxy_username",
            "proxy_password",
            "username",
            "password",
            "pulp_labels",
            "pulp_last_updated",
            "download_concurrency",
            "max_retries",
            "policy",
            "total_timeout",
            "connect_timeout",
            "sock_connect_timeout",
            "sock_read_timeout",
            "headers",
            "rate_limit",
        )


class RepositorySyncURLSerializer(ValidateFieldsMixin, serializers.Serializer):
    remote = DetailRelatedField(
        required=False,
        view_name_pattern=r"remotes(-.*/.*)-detail",
        queryset=models.Remote.objects.all(),
        help_text=_("A remote to sync from. This will override a remote set on repository."),
    )

    mirror = fields.BooleanField(
        required=False,
        default=False,
        help_text=_(
            "If ``True``, synchronization will remove all content that is not present in "
            "the remote repository. If ``False``, sync will be additive only."
        ),
    )

    def validate(self, data):
        data = super().validate(data)

        try:
            remote = models.Repository.objects.get(pk=self.context["repository_pk"]).remote
        except KeyError:
            remote = None

        if "remote" not in data and not remote:
            raise serializers.ValidationError(
                {"remote": _("This field is required since a remote is not set on the repository.")}
            )

        return data


class ContentSummarySerializer(serializers.Serializer):
    """
    Serializer for the RepositoryVersion content summary
    """

    def to_representation(self, obj):
        """
        The summary of contained content.

        Returns:
            dict: The dictionary has the following format.::

                {
                    'added': {<pulp_type>: {'count': <count>, 'href': <href>},
                    'removed': {<pulp_type>: {'count': <count>, 'href': <href>},
                    'present': {<pulp_type>: {'count': <count>, 'href': <href>},
                }

        """
        to_return = {"added": {}, "removed": {}, "present": {}}
        for count_detail in obj.counts.all():
            count_type = count_detail.get_count_type_display()
            item_dict = {"count": count_detail.count, "href": count_detail.content_href}
            to_return[count_type][count_detail.content_type] = item_dict

        return to_return

    def to_internal_value(self, data):
        """
        Setting the internal value.
        """
        return {
            self.added: data["added"],
            self.removed: data["removed"],
            self.present: data["present"],
        }

    added = serializers.DictField(child=serializers.DictField())

    removed = serializers.DictField(child=serializers.DictField())

    present = serializers.DictField(child=serializers.DictField())


class RepositoryVersionSerializer(ModelSerializer, NestedHyperlinkedModelSerializer):
    pulp_href = RepositoryVersionIdentityField()
    number = serializers.IntegerField(read_only=True)
    repository = DetailRelatedField(
        view_name_pattern=r"repositories(-.*/.*)?-detail",
        read_only=True,
    )
    base_version = RepositoryVersionRelatedField(
        required=False,
        help_text=_(
            "A repository version whose content was used as the initial set of content "
            "for this repository version"
        ),
    )
    content_summary = ContentSummarySerializer(
        help_text=_(
            "Various count summaries of the content in the version and the HREF to view them."
        ),
        source="*",
        read_only=True,
    )

    class Meta:
        model = models.RepositoryVersion
        fields = ModelSerializer.Meta.fields + (
            "pulp_href",
            "number",
            "repository",
            "base_version",
            "content_summary",
        )


class RepositoryAddRemoveContentSerializer(ModelSerializer, NestedHyperlinkedModelSerializer):
    add_content_units = serializers.ListField(
        help_text=_(
            "A list of content units to add to a new repository version. This content is "
            "added after remove_content_units are removed."
        ),
        required=False,
    )
    remove_content_units = serializers.ListField(
        help_text=_(
            "A list of content units to remove from the latest repository version. "
            "You may also specify '*' as an entry to remove all content. This content is "
            "removed before add_content_units are added."
        ),
        required=False,
    )
    base_version = RepositoryVersionRelatedField(
        required=False,
        help_text=_(
            "A repository version whose content will be used as the initial set of content "
            "for the new repository version"
        ),
    )

    def validate_remove_content_units(self, value):
        if len(value) > 1 and "*" in value:
            raise serializers.ValidationError("Cannot supply content units and '*'.")
        return value

    class Meta:
        model = models.RepositoryVersion
        fields = ["add_content_units", "remove_content_units", "base_version"]



from gettext import gettext as _

import asyncio

from pulpcore.constants import ALL_KNOWN_CONTENT_CHECKSUMS
from pulpcore.plugin.models import Artifact


class DeclarativeArtifact:
    """
    Relates an :class:`~pulpcore.plugin.models.Artifact`, how to download it, and its
    `relative_path` used later during publishing.

    This is used by the Stages API stages to determine if an
    :class:`~pulpcore.plugin.models.Artifact` is already present and ensure Pulp can download it in
    the future. The `artifact` can be either saved or unsaved. If unsaved, the `artifact` attributes
    may be incomplete because not all digest information can be computed until the
    :class:`~pulpcore.plugin.models.Artifact` is downloaded.

    Attributes:
        artifact (:class:`~pulpcore.plugin.models.Artifact`): An
            :class:`~pulpcore.plugin.models.Artifact` either saved or unsaved. If unsaved, it
            may have partial digest information attached to it.
        url (str): the url to fetch the :class:`~pulpcore.plugin.models.Artifact` from.
        relative_path (str): the relative_path this :class:`~pulpcore.plugin.models.Artifact`
            should be published at for any Publication.
        remote (:class:`~pulpcore.plugin.models.Remote`): The remote used to fetch this
            :class:`~pulpcore.plugin.models.Artifact`.
        extra_data (dict): A dictionary available for additional data to be stored in.
        deferred_download (bool): Whether this artifact should be downloaded and saved
            in the artifact stages. Defaults to `False`. See :ref:`on-demand-support`.

    Raises:
        ValueError: If `artifact`, `url`, or `relative_path` are not specified. If `remote` is not
        specified and `artifact` doesn't have a file.
    """

    __slots__ = ("artifact", "url", "relative_path", "remote", "extra_data", "deferred_download")

    def __init__(
        self,
        artifact=None,
        url=None,
        relative_path=None,
        remote=None,
        extra_data=None,
        deferred_download=False,
    ):
        if not url:
            raise ValueError(_("DeclarativeArtifact must have a 'url'"))
        if not relative_path:
            raise ValueError(_("DeclarativeArtifact must have a 'relative_path'"))
        if not artifact:
            raise ValueError(_("DeclarativeArtifact must have a 'artifact'"))
        if not remote and not artifact.file:
            raise ValueError(
                _(
                    "DeclarativeArtifact must have a 'remote' if the Artifact doesn't "
                    "have a file backing it."
                )
            )
        self.artifact = artifact
        self.url = url
        self.relative_path = relative_path
        self.remote = remote
        self.extra_data = extra_data or {}
        self.deferred_download = deferred_download

    async def download(self):
        """
        Download content and update the associated Artifact.

        Returns:
            Returns the :class:`~pulpcore.plugin.download.DownloadResult` of the Artifact.
        """
        expected_digests = {}
        validation_kwargs = {}
        for digest_name in ALL_KNOWN_CONTENT_CHECKSUMS:
            digest_value = getattr(self.artifact, digest_name)
            if digest_value:
                expected_digests[digest_name] = digest_value
        if expected_digests:
            validation_kwargs["expected_digests"] = expected_digests
        if self.artifact.size:
            expected_size = self.artifact.size
            validation_kwargs["expected_size"] = expected_size
        downloader = self.remote.get_downloader(url=self.url, **validation_kwargs)
        # Custom downloaders may need extra information to complete the request.
        download_result = await downloader.run(extra_data=self.extra_data)
        self.artifact = Artifact(**download_result.artifact_attributes, file=download_result.path)
        return download_result


class DeclarativeContent:
    """
    Relates a Content unit and zero or more :class:`~pulpcore.plugin.stages.DeclarativeArtifact`
    objects.

    This is used by the Stages API stages to determine if a Content unit is already present and
    ensure all of its associated :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects are
    related correctly. The `content` can be either saved or unsaved depending on where in the Stages
    API pipeline this is used.

    Attributes:
        content (subclass of :class:`~pulpcore.plugin.models.Content`): A Content unit, possibly
            unsaved
        d_artifacts (list): A list of zero or more
            :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects associated with `content`.
        extra_data (dict): A dictionary available for additional data to be stored in.

    Raises:
        ValueError: If `content` is not specified.
    """

    __slots__ = (
        "content",
        "d_artifacts",
        "extra_data",
        "_future",
        "_thaw_queue_event",
        "_resolved",
    )

    def __init__(self, content=None, d_artifacts=None, extra_data=None):
        if not content:
            raise ValueError(_("DeclarativeContent must have a 'content'"))
        self.content = content
        self.d_artifacts = d_artifacts or []
        self.extra_data = extra_data or {}
        self._future = None
        self._thaw_queue_event = None
        self._resolved = False

    @property
    def does_batch(self):
        """Whether this content is being awaited on and must therefore not wait forever in batches.
        When overwritten in subclasses, a `True` value must never be turned into `False`.
        """
        return not self._resolved and self._future is None

    async def resolution(self):
        """Coroutine that waits for the content to be saved to database.
        Returns the content unit."""
        if self._resolved:
            # Already resolved ~> shortcut
            return self.content
        if self._future is None:
            # We do not yet have a future
            self._future = asyncio.get_event_loop().create_future()
            if self._thaw_queue_event:
                # We have a future now but are still stuck in a queue
                self._thaw_queue_event.set()
        # Now we wait
        return await self._future

    def resolve(self):
        """Resolve this content unit and notify any waiting tasks."""
        self._resolved = True
        if self._future:
            self._future.set_result(self.content)
            self._future = None

    def __str__(self):
        return str(self.content.__class__.__name__)



# Generated by Django 2.2.16 on 2020-09-25 22:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0051_timeoutfields'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='logging_cid',
            field=models.CharField(db_index=True, default='', max_length=256),
        ),
    ]



"""Tests that perform action over remotes"""

import unittest

from pulp_smash import api, config
from pulp_smash.pulp3.utils import gen_repo, get_content, sync

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_NAME,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import gen_file_remote
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa


class RemotesTestCase(unittest.TestCase):
    """Verify remotes can be used with different repos."""

    def test_all(self):
        """Verify remotes can be used with different repos.

        This test explores the design choice stated in `Pulp #3341`_ that
        remove the FK from remotes to repository.
        Allowing remotes to be used with different
        repositories.

        .. _Pulp #3341: https://pulp.plan.io/issues/3341

        Do the following:

        1. Create a remote.
        2. Create 2 repositories.
        3. Sync both repositories using the same remote.
        4. Assert that the two repositories have the same contents.
        """
        cfg = config.get_config()

        # Create a remote.
        client = api.Client(cfg, api.json_handler)
        body = gen_file_remote()
        remote = client.post(FILE_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["pulp_href"])

        # Create and sync repos.
        repos = []
        for _ in range(2):
            repo = client.post(FILE_REPO_PATH, gen_repo())
            self.addCleanup(client.delete, repo["pulp_href"])
            sync(cfg, remote, repo)
            repos.append(client.get(repo["pulp_href"]))

        # Compare contents of repositories.
        contents = []
        for repo in repos:
            contents.append(get_content(repo)[FILE_CONTENT_NAME])
        self.assertEqual(
            {content["pulp_href"] for content in contents[0]},
            {content["pulp_href"] for content in contents[1]},
        )



from gettext import gettext as _
from logging import getLogger
from tempfile import NamedTemporaryFile

from django.core.files import File

from pulpcore.app import files, models
from pulpcore.app.models import CreatedResource
from pulpcore.app.serializers import ArtifactSerializer

log = getLogger(__name__)


def commit(upload_id, sha256):
    """
    Commit the upload and turn it into an artifact.

    Args:
        upload_id (int): The upload primary key
        sha256 (str): The checksum for the uploaded file
    """
    try:
        upload = models.Upload.objects.get(pk=upload_id)
    except models.Upload.DoesNotExist:
        log.info(_("The upload was not found. Nothing to do."))
        return

    chunks = models.UploadChunk.objects.filter(upload=upload).order_by("offset")
    with NamedTemporaryFile("ab") as temp_file:
        for chunk in chunks:
            temp_file.write(chunk.file.read())
        temp_file.flush()

        file = files.PulpTemporaryUploadedFile.from_file(File(open(temp_file.name, "rb")))

    data = {"file": file, "sha256": sha256}
    serializer = ArtifactSerializer(data=data)
    serializer.is_valid(raise_exception=True)
    artifact = serializer.save()

    resource = CreatedResource(content_object=artifact)
    resource.save()

    # delete the upload since it can't be reused to create another artifact
    upload.delete()



import unittest


from pulp_smash import api, cli, config

from pulpcore.client.pulpcore import ApiClient, GroupsApi, UsersApi


class UserTestCase(unittest.TestCase):
    """Test REST API for users."""

    @classmethod
    def setUpClass(cls):
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.client = ApiClient(
            configuration=cls.cfg.get_bindings_config(),
        )
        cls.cli_client = cli.Client(cls.cfg)

        cls.user_api = UsersApi(cls.client)

    def test_list_users(self):
        """Test that users can be listed."""
        users = self.user_api.list()
        self.assertIn("admin", [user.username for user in users.results])

    def test_filter_users(self):
        """Test that users can be filterd."""
        users = self.user_api.list(username="admin")
        self.assertEqual(len(users.results), 1)


class GroupTestCase(unittest.TestCase):
    """Test REST API for users."""

    @classmethod
    def setUpClass(cls):
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)
        cls.client = ApiClient(
            configuration=cls.cfg.get_bindings_config(),
        )
        cls.cli_client = cli.Client(cls.cfg)

        cls.group_api = GroupsApi(cls.client)

    def test_list_groups(self):
        """Test that a group can be crud."""

        groups = self.group_api.list()
        self.assertNotIn("test_newbees", [group.name for group in groups.results])
        group_href = self.group_api.create(group={"name": "test_newbees"}).pulp_href
        try:
            groups = self.group_api.list()
            self.assertIn("test_newbees", [group.name for group in groups.results])
        finally:
            self.group_api.delete(group_href)
        groups = self.group_api.list()
        self.assertNotIn("test_newbees", [group.name for group in groups.results])

    def test_filter_users(self):
        """Test that groups can be filterd."""

        groups = self.group_api.list(name__contains="test_")
        self.assertEqual(len(groups.results), 0)

        group_hrefs = [
            self.group_api.create(group={"name": name}).pulp_href
            for name in ["test_newbees", "test_admins"]
        ]
        try:
            groups = self.group_api.list(name__contains="test_")
            self.assertEqual(len(groups.results), 2)
            groups = self.group_api.list(name__contains="test_new")
            self.assertEqual(len(groups.results), 1)
            groups = self.group_api.list(name="test_newbees")
            self.assertEqual(len(groups.results), 1)
        finally:
            for group_href in group_hrefs:
                self.group_api.delete(group_href)



from types import SimpleNamespace

TASKING_CONSTANTS = SimpleNamespace(
    # The name of resource manager entries in the workers table
    RESOURCE_MANAGER_WORKER_NAME="resource-manager",
    # The amount of time (in seconds) between checks
    JOB_MONITORING_INTERVAL=5,
)



from django_filters.rest_framework import filters
from rest_framework import mixins

from pulpcore.app.models import AccessPolicy
from pulpcore.app.serializers import AccessPolicySerializer
from pulpcore.app.viewsets import BaseFilterSet, NamedModelViewSet
from pulpcore.app.viewsets.base import NAME_FILTER_OPTIONS


class AccessPolicyFilter(BaseFilterSet):
    """
    FilterSet for AccessPolicy.
    """

    customized = filters.BooleanFilter()

    class Meta:
        model = AccessPolicy
        fields = {"viewset_name": NAME_FILTER_OPTIONS, "customized": ["exact"]}


class AccessPolicyViewSet(
    NamedModelViewSet, mixins.UpdateModelMixin, mixins.RetrieveModelMixin, mixins.ListModelMixin
):
    """
    ViewSet for AccessPolicy.

    NOTE: This API endpoint is in "tech preview" and subject to change

    """

    queryset = AccessPolicy.objects.all()
    serializer_class = AccessPolicySerializer
    endpoint_name = "access_policies"
    ordering = "viewset_name"
    filterset_class = AccessPolicyFilter



"""Tests related to content cache."""
import requests
import unittest
from urllib.parse import urljoin

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_distribution, gen_repo

from .constants import PULP_CONTENT_BASE_URL
from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)
from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositoryAddRemoveContent,
    RepositorySyncURL,
    RepositoriesFileApi,
    RemotesFileApi,
    PublicationsFileApi,
    FileFilePublication,
    DistributionsFileApi,
    PatchedfileFileDistribution,
)
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class ContentCacheTestCache(unittest.TestCase):
    """Test content cache"""

    @classmethod
    def setUpClass(cls):
        """Sets up class"""
        client = gen_file_client()
        cls.cont_api = ContentFilesApi(client)
        cls.repo_api = RepositoriesFileApi(client)
        cls.remote_api = RemotesFileApi(client)
        cls.pub_api = PublicationsFileApi(client)
        cls.dis_api = DistributionsFileApi(client)
        cls.repo = cls.repo_api.create(gen_repo(autopublish=True))
        cls.remote = cls.remote_api.create(gen_file_remote())
        body = RepositorySyncURL(remote=cls.remote.pulp_href)
        created = monitor_task(cls.repo_api.sync(cls.repo.pulp_href, body).task).created_resources
        cls.repo = cls.repo_api.read(cls.repo.pulp_href)
        cls.pub1 = cls.pub_api.read(created[1])
        body = FileFilePublication(repository=cls.repo.pulp_href)
        cls.pub2 = cls.pub_api.read(
            monitor_task(cls.pub_api.create(body).task).created_resources[0]
        )
        cls.pub3 = []
        response = cls.dis_api.create(gen_distribution(repository=cls.repo.pulp_href))
        cls.distro = cls.dis_api.read(monitor_task(response.task).created_resources[0])
        cls.distro2 = []
        cls.url = urljoin(PULP_CONTENT_BASE_URL, f"{cls.distro.base_path}/")

    @classmethod
    def tearDownClass(cls):
        """Tears the class down"""
        cls.remote_api.delete(cls.remote.pulp_href)
        cls.dis_api.delete(cls.distro.pulp_href)

    def test_01_basic_cache_access(self):
        """Checks responses are cached for content"""
        files = ["", "", "PULP_MANIFEST", "PULP_MANIFEST", "1.iso", "1.iso"]
        for i, file in enumerate(files):
            self.assertEqual((200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file), file)

    def test_02_remove_repository_invalidates(self):
        """Checks removing repository from distribution invalidates the cache"""
        body = PatchedfileFileDistribution(repository="")
        monitor_task(self.dis_api.partial_update(self.distro.pulp_href, body).task)
        files = ["", "PULP_MANIFEST", "1.iso"]
        for file in files:
            self.assertEqual((404, None), self.check_cache(file), file)

    def test_03_restore_repository(self):
        """Checks that responses are cacheable after repository is added back"""
        body = PatchedfileFileDistribution(repository=self.repo.pulp_href)
        monitor_task(self.dis_api.partial_update(self.distro.pulp_href, body).task)
        files = ["", "", "PULP_MANIFEST", "PULP_MANIFEST", "1.iso", "1.iso"]
        for i, file in enumerate(files):
            self.assertEqual((200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file), file)

    def test_04_multiple_distributions(self):
        """Add a new distribution and check that its responses are cached separately"""
        response = self.dis_api.create(gen_distribution(repository=self.repo.pulp_href))
        self.distro2.append(self.dis_api.read(monitor_task(response.task).created_resources[0]))
        url = urljoin(PULP_CONTENT_BASE_URL, f"{self.distro2[0].base_path}/")
        files = ["", "", "PULP_MANIFEST", "PULP_MANIFEST", "1.iso", "1.iso"]
        for i, file in enumerate(files):
            self.assertEqual(
                (200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file, url), file
            )

    def test_05_invalidate_multiple_distributions(self):
        """Test that updating a repository pointed by multiple distributions invalidates all"""
        url = urljoin(PULP_CONTENT_BASE_URL, f"{self.distro2[0].base_path}/")
        cfile = self.cont_api.list(
            relative_path="1.iso", repository_version=self.repo.latest_version_href
        ).results[0]
        body = RepositoryAddRemoveContent(remove_content_units=[cfile.pulp_href])
        response = monitor_task(self.repo_api.modify(self.repo.pulp_href, body).task)
        self.pub3.append(self.pub_api.read(response.created_resources[1]))
        files = ["", "", "PULP_MANIFEST", "PULP_MANIFEST", "2.iso", "2.iso"]
        for i, file in enumerate(files):
            self.assertEqual((200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file), file)
            self.assertEqual(
                (200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file, url), file
            )

    def test_06_delete_distribution_invalidates_one(self):
        """Tests that deleting one distribution sharing a repository only invalidates its cache"""
        url = urljoin(PULP_CONTENT_BASE_URL, f"{self.distro2[0].base_path}/")
        monitor_task(self.dis_api.delete(self.distro2[0].pulp_href).task)
        files = ["", "PULP_MANIFEST", "2.iso"]
        for file in files:
            self.assertEqual((200, "HIT"), self.check_cache(file), file)
            self.assertEqual((404, None), self.check_cache(file, url), file)

    def test_07_delete_extra_pub_doesnt_invalidate(self):
        """Test that deleting a publication not being served doesn't invalidate cache"""
        self.pub_api.delete(self.pub2.pulp_href)
        files = ["", "PULP_MANIFEST", "2.iso"]
        for file in files:
            self.assertEqual((200, "HIT"), self.check_cache(file), file)

    def test_08_delete_served_pub_does_invalidate(self):
        """Test that deleting the serving publication does invalidate the cache"""
        # Reverts back to serving self.pub1
        self.pub_api.delete(self.pub3[0].pulp_href)
        files = ["", "", "PULP_MANIFEST", "PULP_MANIFEST", "2.iso", "2.iso"]
        for i, file in enumerate(files):
            self.assertEqual((200, "HIT" if i % 2 == 1 else "MISS"), self.check_cache(file), file)

    def test_09_delete_repo_invalidates(self):
        """Tests that deleting a repository invalidates the cache"""
        monitor_task(self.repo_api.delete(self.repo.pulp_href).task)
        files = ["", "PULP_MANIFEST", "2.iso"]
        for file in files:
            self.assertEqual((404, None), self.check_cache(file), file)

    def test_10_no_error_when_accessing_invalid_file(self):
        """Tests that accessing a file that doesn't exist on content app gives 404"""
        files = ["invalid", "another/bad-one", "DNE/"]
        url = PULP_CONTENT_BASE_URL
        for file in files:
            self.assertEqual((404, None), self.check_cache(file, url=url), file)

    def check_cache(self, file, url=None):
        """Helper to check if cache miss or hit"""
        url = urljoin(url or self.url, file)
        r = requests.get(url)
        if r.history:
            r = r.history[0]
            return 200 if r.status_code == 302 else r.status_code, r.headers.get("X-PULP-CACHE")
        return r.status_code, r.headers.get("X-PULP-CACHE")



# Generated by Django 2.2.11 on 2020-03-30 16:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0024_use_local_storage_for_uploads'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='parent_task',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='child_tasks', to='core.Task'),
        ),
    ]



from unittest import TestCase

import mock
from pulpcore.app.models import Artifact
from pulpcore.app.serializers import ArtifactSerializer
from pulpcore.constants import ALL_KNOWN_CONTENT_CHECKSUMS
from rest_framework import serializers


class TestArtifactSerializer(TestCase):
    def test_validate_file_checksum(self):
        mock_file = mock.MagicMock(size=42)
        mock_file.hashers.__getitem__.return_value.hexdigest.return_value = "asdf"

        data = {"file": mock_file}
        serializer = ArtifactSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        new_data = serializer.validated_data
        self.assertEqual(new_data["file"], mock_file)
        self.assertEqual(new_data["size"], 42)
        for csum in Artifact.DIGEST_FIELDS:
            self.assertEqual(new_data[csum], "asdf")

        for csum in ALL_KNOWN_CONTENT_CHECKSUMS.difference(Artifact.DIGEST_FIELDS):
            self.assertFalse(csum in new_data, f"Found forbidden checksum {csum}")

        # This part of the test will only fire if the system-under-test has forbidden
        # use of 'md5'
        if "md5" not in Artifact.DIGEST_FIELDS:
            data = {"file": mock_file, "md5": "asdf"}
            with self.assertRaises(serializers.ValidationError) as cm:  # noqa
                serializer.validate(data)

    def test_emtpy_data(self):
        data = {}
        serializer = ArtifactSerializer(data=data)
        self.assertFalse(serializer.is_valid())



from gettext import gettext as _
import uuid

from django.contrib.contenttypes.fields import GenericRelation, GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import options
from django.db.models.base import ModelBase
from django_lifecycle import LifecycleModel


class Label(LifecycleModel):
    """Model for handling resource labels.

    Labels are key/value data that can be associated with any BaseModel.

    Fields:
        pulp_id (models.UUIDField): Primary key identifier
        object_id (models.UUIDField): Resource id
        key (models.CharField): Key of the label
        value (models.TextField): Value of the label

    Relations:
        content_object (GenericForeignKey): Associated resource
        content_type (models.ForeignKey): Content type of the resource
    """

    pulp_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.UUIDField()
    key = models.CharField(max_length=200, db_index=True)
    value = models.TextField(null=True, db_index=True)

    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        unique_together = [["content_type", "object_id", "key"]]


class BaseModel(LifecycleModel):
    """Base model class for all Pulp models.

    This model inherits from `LifecycleModel` which allows all Pulp models to be used with
    `django-lifecycle`.

    Fields:
        pulp_created (models.DateTimeField): Created timestamp UTC.
        pulp_last_updated (models.DateTimeField): Last updated timestamp UTC.

    Relations:
        pulp_labels (GenericRelation): A list of key/value labels.

    References:

        * https://docs.djangoproject.com/en/1.8/topics/db/models/#automatic-primary-key-fields
        * https://rsinger86.github.io/django-lifecycle/

    """

    pulp_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    pulp_created = models.DateTimeField(auto_now_add=True)
    pulp_last_updated = models.DateTimeField(auto_now=True, null=True)
    pulp_labels = GenericRelation(Label)

    class Meta:
        abstract = True

    def __str__(self):
        try:
            # if we have a name, use it
            return "<{}: {}>".format(self._meta.object_name, self.name)
        except AttributeError:
            # if we don't, use the pk
            return "<{}: pk={}>".format(self._meta.object_name, self.pk)

    def __repr__(self):
        return str(self)


class MasterModelMeta(ModelBase):
    def __new__(cls, name, bases, attrs, **kwargs):
        """Override __new__ to set the default_related_name."""
        if BaseModel not in bases and MasterModel not in bases:  # Only affects "Detail" models.
            meta = attrs.get("Meta")
            default_related_name = getattr(meta, "default_related_name", None)
            abstract = getattr(meta, "abstract", None)

            if not default_related_name and not abstract:
                raise Exception(
                    _("The 'default_related_name' option has not been set for {class_name}").format(
                        class_name=name
                    )
                )

        new_class = super().__new__(cls, name, bases, attrs, **kwargs)
        return new_class


class MasterModel(BaseModel, metaclass=MasterModelMeta):
    """Base model for the "Master" model in a "Master-Detail" relationship.

    Provides methods for casting down to detail types, back up to the master type,
    as well as a model field for tracking the type.

    Attributes:

        TYPE (str): Default constant value saved into the ``pulp_type``
            field of Model instances

    Fields:

        pulp_type: The user-facing string identifying the detail type of this model

    Warning:
        Subclasses of this class rely on there being no other parent/child Model
        relationships than the Master/Detail relationship. All subclasses must use
        only abstract Model base classes for MasterModel to behave properly.
        Specifically, OneToOneField relationships must not be used in any MasterModel
        subclass.

    """

    # TYPE is the user-facing string that describes this type. It is used to construct API
    # endpoints for Detail models, and will be seen in the URLs generated for those Detail models.
    # It can also be used for filtering across a relation where a model is related to a Master
    # model. Set this to something reasonable in Master and Detail model classes, e.g. when
    # create a master model, like "Remote", its TYPE value could be "remote". Then, when
    # creating a Remote Detail class like PackageRemote, its pulp_type value could be "package",
    # not "package_remote", since "package_remote" would be redundant in the context of
    # a remote Master model.
    TYPE = None

    # This field must have a value when models are saved, and defaults to the value of
    # the TYPE attribute on the Model being saved (seen above).
    pulp_type = models.TextField(null=False, default=None, db_index=True)

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        # instances of "detail" models that subclass MasterModel are exposed
        # on instances of MasterModel by the string stored in that model's TYPE attr.
        # Storing this pulp_type in a column on the MasterModel next to makes it trivial
        # to filter for specific detail model types across master's relations.
        # Prepend the TYPE defined on a detail model with a django app label.
        # If a plugin sets the type field themselves, it's used as-is.
        if not self.pulp_type:
            self.pulp_type = self.get_pulp_type()
        return super().save(*args, **kwargs)

    @classmethod
    def get_pulp_type(cls):
        """Get the "pulp_type" string associated with this MasterModel type."""
        return "{app_label}.{type}".format(app_label=cls._meta.app_label, type=cls.TYPE)

    def cast(self):
        """Return a "Detail" model instance of this master-detail pair.

        If this model is already an instance of its detail type, it will return itself.
        """
        # Go through our related objects, find the one that's a subclass of this model
        # on a OneToOneField, which identifies it as a potential detail relation.
        for rel in self._meta.related_objects:
            if rel.one_to_one and issubclass(rel.related_model, self._meta.model):
                # The name of this relation is the name of the attr on the model instance.
                # If that attr as a value, that means a row exists for this model in the
                # related detail table. Cast and return this value, recursively following
                # master/detail relationships down to the last table (the most detailed).
                try:
                    return getattr(self, rel.name).cast()
                except AttributeError:
                    continue
        else:
            # The for loop exited normally, there are no more detailed models than this
            # one in this instance's master/detail ancestry, so return here.
            return self

    @property
    def master(self):
        """The "Master" model instance of this master-detail pair

        If this is already the master model instance, it will return itself.
        """
        if self._meta.master_model:
            return self._meta.master_model(pk=self.pk)
        else:
            return self

    def __str__(self):
        # similar to Model's __str__, but type-aware
        cast = self.cast()
        if cast is self:
            return super().__str__()

        try:
            return "<{} (pulp_type={}): {}>".format(
                self._meta.object_name, cast.pulp_type, cast.name
            )
        except AttributeError:
            return "<{} (pulp_type={}): pk={}>".format(
                self._meta.object_name, cast.pulp_type, cast.pk
            )


# Add properties to model _meta info to support master/detail models
# If this property is not None on a Model, then that Model is a Detail Model.
# Doing this in a non-monkeypatch way would mean a lot of effort to achieve the same result
# (e.g. custom model metaclass, custom Options implementation, etc). These could be classmethods
# on Model classes, but it's easy enough to use the model's _meta namespace to do this, since
# that's where other methods like this exist in Django.
def master_model(options):
    """
    The Master model class of this Model's Master/Detail relationship.

    Accessible at ``<model_class>._meta.master_model``, the Master model class in a Master/Detail
    relationship is the most generic non-abstract Model in this model's multiple-table chain
    of inheritance.

    If this model is not a detail model, None will be returned.
    """
    # If this isn't even a MasterModel descendant, don't bother.
    if not issubclass(options.model, MasterModel):
        return None
    try:
        # The last item in this list is the oldest ancestor. Since the MasterModel usage
        # is to declare your master by subclassing MasterModel, and MasterModel is abstract,
        # the oldest ancestor model is the Master Model.
        return options.get_parent_list()[-1]
    except IndexError:
        # Also None if this model is itself the master.
        return None


options.Options.master_model = property(master_model)



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.app import models
from pulpcore.app.serializers import (
    GroupProgressReportSerializer,
    IdentityField,
    ModelSerializer,
    ProgressReportSerializer,
    RelatedField,
    RelatedResourceField,
    TaskGroupStatusCountField,
)
from pulpcore.constants import TASK_STATES


class CreatedResourceSerializer(RelatedResourceField):
    class Meta:
        model = models.CreatedResource
        fields = []


class TaskSerializer(ModelSerializer):
    pulp_href = IdentityField(view_name="tasks-detail")
    state = serializers.CharField(
        help_text=_(
            "The current state of the task. The possible values include:"
            " 'waiting', 'skipped', 'running', 'completed', 'failed', 'canceled' and 'canceling'."
        ),
        read_only=True,
    )
    name = serializers.CharField(help_text=_("The name of task."))
    logging_cid = serializers.CharField(
        help_text=_("The logging correlation id associated with this task")
    )
    started_at = serializers.DateTimeField(
        help_text=_("Timestamp of the when this task started execution."), read_only=True
    )
    finished_at = serializers.DateTimeField(
        help_text=_("Timestamp of the when this task stopped execution."), read_only=True
    )
    error = serializers.DictField(
        child=serializers.JSONField(),
        help_text=_(
            "A JSON Object of a fatal error encountered during the execution of this task."
        ),
        read_only=True,
    )
    worker = RelatedField(
        help_text=_(
            "The worker associated with this task."
            " This field is empty if a worker is not yet assigned."
        ),
        read_only=True,
        view_name="workers-detail",
    )
    parent_task = RelatedField(
        help_text=_("The parent task that spawned this task."),
        read_only=True,
        view_name="tasks-detail",
    )
    child_tasks = RelatedField(
        help_text=_("Any tasks spawned by this task."),
        many=True,
        read_only=True,
        view_name="tasks-detail",
    )
    task_group = RelatedField(
        help_text=_("The task group that this task is a member of."),
        read_only=True,
        view_name="task-groups-detail",
    )
    progress_reports = ProgressReportSerializer(many=True, read_only=True)
    created_resources = CreatedResourceSerializer(
        help_text=_("Resources created by this task."),
        many=True,
        read_only=True,
        view_name="None",  # This is a polymorphic field. The serializer does not need a view name.
    )
    reserved_resources_record = serializers.ListField(
        child=serializers.CharField(),
        help_text=_("A list of resources required by that task."),
        read_only=True,
    )

    class Meta:
        model = models.Task
        fields = ModelSerializer.Meta.fields + (
            "state",
            "name",
            "logging_cid",
            "started_at",
            "finished_at",
            "error",
            "worker",
            "parent_task",
            "child_tasks",
            "task_group",
            "progress_reports",
            "created_resources",
            "reserved_resources_record",
        )


class MinimalTaskSerializer(TaskSerializer):
    class Meta:
        model = models.Task
        fields = ModelSerializer.Meta.fields + (
            "name",
            "state",
            "started_at",
            "finished_at",
            "worker",
        )


class TaskGroupSerializer(ModelSerializer):
    pulp_href = IdentityField(view_name="task-groups-detail")
    description = serializers.CharField(help_text=_("A description of the task group."))
    all_tasks_dispatched = serializers.BooleanField(
        help_text=_("Whether all tasks have been spawned for this task group.")
    )

    waiting = TaskGroupStatusCountField(
        state=TASK_STATES.WAITING, help_text=_("Number of tasks in the 'waiting' state")
    )
    skipped = TaskGroupStatusCountField(
        state=TASK_STATES.SKIPPED, help_text=_("Number of tasks in the 'skipped' state")
    )
    running = TaskGroupStatusCountField(
        state=TASK_STATES.RUNNING, help_text=_("Number of tasks in the 'running' state")
    )
    completed = TaskGroupStatusCountField(
        state=TASK_STATES.COMPLETED, help_text=_("Number of tasks in the 'completed' state")
    )
    canceled = TaskGroupStatusCountField(
        state=TASK_STATES.CANCELED, help_text=_("Number of tasks in the 'canceled' state")
    )
    failed = TaskGroupStatusCountField(
        state=TASK_STATES.FAILED, help_text=_("Number of tasks in the 'failed' state")
    )
    canceling = TaskGroupStatusCountField(
        state=TASK_STATES.CANCELING, help_text=_("Number of tasks in the 'canceling' state")
    )
    group_progress_reports = GroupProgressReportSerializer(many=True, read_only=True)
    tasks = MinimalTaskSerializer(many=True, read_only=True)

    class Meta:
        model = models.TaskGroup
        fields = (
            "pulp_href",
            "description",
            "all_tasks_dispatched",
            "waiting",
            "skipped",
            "running",
            "completed",
            "canceled",
            "failed",
            "canceling",
            "group_progress_reports",
            "tasks",
        )


class TaskCancelSerializer(ModelSerializer):
    state = serializers.CharField(
        help_text=_("The desired state of the task. Only 'canceled' is accepted."),
    )

    class Meta:
        model = models.Task
        fields = ("state",)


class ContentAppStatusSerializer(ModelSerializer):
    name = serializers.CharField(help_text=_("The name of the worker."), read_only=True)
    last_heartbeat = serializers.DateTimeField(
        help_text=_("Timestamp of the last time the worker talked to the service."), read_only=True
    )

    class Meta:
        model = models.ContentAppStatus
        fields = ("name", "last_heartbeat")


class WorkerSerializer(ModelSerializer):
    pulp_href = IdentityField(view_name="workers-detail")

    name = serializers.CharField(help_text=_("The name of the worker."), read_only=True)
    last_heartbeat = serializers.DateTimeField(
        help_text=_("Timestamp of the last time the worker talked to the service."), read_only=True
    )
    # disable "created" because we don't care about it
    created = None

    class Meta:
        model = models.Worker
        _base_fields = tuple(set(ModelSerializer.Meta.fields) - set(["created"]))
        fields = _base_fields + ("name", "last_heartbeat")



# Generated by Django 3.2.6 on 2021-08-25 20:10

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.access_policy


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0074_acs'),
    ]

    operations = [
        migrations.CreateModel(
            name='RBACContentGuard',
            fields=[
                ('contentguard_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_rbaccontentguard', serialize=False, to='core.contentguard')),
            ],
            options={
                'permissions': (('download_rbaccontentguard', 'Can Download Content'),),
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.contentguard', pulpcore.app.models.access_policy.AutoAddObjPermsMixin, pulpcore.app.models.access_policy.AutoDeleteObjPermsMixin),
        ),
    ]



"""Utilities for pulpcore API tests."""
import string
from datetime import datetime
from random import choice


def gen_username(length=10, valid_characters=True):
    """Generate username given a certain length or punctuation to be used."""
    valid_punctuation = "@.+-_"
    if valid_characters:
        return "".join(
            choice(string.ascii_letters + string.digits + valid_punctuation) for _ in range(length)
        )
    invalid_puntuation = "".join(
        value for value in string.punctuation if value not in valid_punctuation
    )
    return "".join(choice(invalid_puntuation) for _ in range(length))


def parse_date_from_string(s, parse_format="%Y-%m-%dT%H:%M:%S.%fZ"):
    """Parse string to datetime object.

    :param s: str like '2018-11-18T21:03:32.493697Z'
    :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
    :return: datetime.datetime
    """
    return datetime.strptime(s, parse_format)



"""Test related to the api docs page."""
import unittest

from pulp_smash import api, config
from pulp_smash.pulp3.constants import API_DOCS_PATH
from requests.exceptions import HTTPError

from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class ApiDocsTestCase(unittest.TestCase):
    """Test whether API auto generated docs are available.

    This test targets the following issue:

    * `Pulp Smash #893 <https://github.com/pulp/pulp-smash/issues/893>`_
    """

    def setUp(self):
        """Create an API Client."""
        cfg = config.get_config()
        self.client = api.Client(cfg)

    def test_valid_credentials(self):
        """Get API documentation with valid credentials.

        Assert the API documentation is returned.
        """
        self.client.get(API_DOCS_PATH)

    def test_no_credentials(self):
        """Get API documentation with no credentials.

        Assert the API documentation is returned.
        """
        del self.client.request_kwargs["auth"]
        self.client.get(API_DOCS_PATH)

    def test_http_method(self):
        """Get API documentation with an HTTP method other than GET.

        Assert an error is returned.
        """
        with self.assertRaises(HTTPError):
            self.client.post(API_DOCS_PATH)



from gettext import gettext as _

from rest_framework import fields, serializers

from pulpcore.app.models import Repository

from pulpcore.app.serializers import RepositoryVersionRelatedField, ValidateFieldsMixin


class ReclaimSpaceSerializer(serializers.Serializer, ValidateFieldsMixin):
    """
    Serializer for reclaim disk space operation.
    """

    repo_hrefs = fields.ListField(
        required=True,
        help_text=_("Will reclaim space for the specified list of repos."),
    )
    repo_versions_keeplist = RepositoryVersionRelatedField(
        help_text=_("Will exclude repo versions from space reclaim."),
        many=True,
        required=False,
    )

    def validate_repo_hrefs(self, value):
        """
        Check that the repo_hrefs is not an empty list and contains all valid hrefs.
        Args:
            value (list): The list supplied by the user
        Returns:
            The list of pks (not hrefs) after validation
        Raises:
            ValidationError: If the list is empty or contains invalid hrefs.
        """
        if len(value) == 0:
            raise serializers.ValidationError("Must not be [].")
        from pulpcore.app.viewsets import NamedModelViewSet

        hrefs_to_return = []
        for href in value:
            hrefs_to_return.append(NamedModelViewSet.get_resource(href, Repository))

        return hrefs_to_return



from gettext import gettext as _
import logging

from django.db.models import Q

from pulpcore.app.files import validate_file_paths
from pulpcore.app.models import Content, ContentArtifact
from pulpcore.app.util import batch_qs


_logger = logging.getLogger(__name__)

__all__ = ["remove_duplicates"]


def remove_duplicates(repository_version):
    """
    Inspect content additions in the `RepositoryVersion` and remove existing repository duplicates.

    This function will inspect the content being added to a repo version and remove any existing
    content which would collide with the content being added to the repository version. It does not
    inspect the content being added for duplicates.

    Some content can have two instances A and B which are unique, but cannot both exist together in
    one repository. For example, pulp_file's content has `relative_path` for that file within the
    repository.

    Any content newly added to the :class:`~pulpcore.plugin.models.RepositoryVersion` is checked
    against existing content in the :class:`~pulpcore.plugin.models.RepositoryVersion` with newer
    "repository duplicates" replace existing "repository duplicates". Each Content model can define
    a `repo_key_fields` attribute with the field names to be compared. If all `repo_key_fields`
    contain the same value for two content units, they are considered "repository duplicates".

    Args:
        repository_version: The :class:`~pulpcore.plugin.models.RepositoryVersion` to be checked
            and possibly modified.
    """
    added_content = repository_version.added(base_version=repository_version.base_version)
    if repository_version.base_version:
        existing_content = repository_version.base_version.content
    else:
        try:
            existing_content = repository_version.previous().content
        except repository_version.DoesNotExist:
            existing_content = Content.objects.none()
    repository = repository_version.repository.cast()
    content_types = {type_obj.get_pulp_type(): type_obj for type_obj in repository.CONTENT_TYPES}

    for pulp_type, type_obj in content_types.items():
        repo_key_fields = type_obj.repo_key_fields
        new_content_qs = type_obj.objects.filter(
            pk__in=added_content.filter(pulp_type=pulp_type)
        ).values(*repo_key_fields)

        if type_obj.repo_key_fields == ():
            continue

        if new_content_qs.count() and existing_content.count():
            _logger.debug(_("Removing duplicates for type: {}".format(type_obj.get_pulp_type())))

            for batch in batch_qs(new_content_qs):
                find_dup_qs = Q()

                for content_dict in batch:
                    item_query = Q(**content_dict)
                    find_dup_qs |= item_query

                duplicates_qs = (
                    type_obj.objects.filter(pk__in=existing_content).filter(find_dup_qs).only("pk")
                )
                repository_version.remove_content(duplicates_qs)


def validate_duplicate_content(version):
    """
    Validate that a repository version doesn't contain duplicate content.

    Uses repo_key_fields to determine if content is duplicated.

    Raises:
        ValueError: If repo version has duplicate content.
    """
    error_messages = []

    for type_obj in version.repository.CONTENT_TYPES:
        if type_obj.repo_key_fields == ():
            continue

        pulp_type = type_obj.get_pulp_type()
        repo_key_fields = type_obj.repo_key_fields
        new_content_total = type_obj.objects.filter(
            pk__in=version.content.filter(pulp_type=pulp_type)
        ).count()
        unique_new_content_total = (
            type_obj.objects.filter(pk__in=version.content.filter(pulp_type=pulp_type))
            .distinct(*repo_key_fields)
            .count()
        )

        if unique_new_content_total < new_content_total:
            error_messages.append(
                _(
                    "More than one {pulp_type} content with the duplicate values for {fields}."
                ).format(pulp_type=pulp_type, fields=", ".join(repo_key_fields))
            )
    if error_messages:
        raise ValueError(
            _("Cannot create repository version. {msg}").format(msg=", ".join(error_messages))
        )


def validate_version_paths(version):
    """
    Validate artifact relative paths for dupes or overlap (e.g. a/b and a/b/c).

    Raises:
        ValueError: If two artifact relative paths overlap
    """
    paths = ContentArtifact.objects.filter(content__pk__in=version.content).values_list(
        "relative_path", flat=True
    )
    try:
        validate_file_paths(paths)
    except ValueError as e:
        raise ValueError(_("Repository version errors : {err}").format(err=e))


def validate_repo_version(version):
    """
    Validate a repo version.

    Checks for duplicate content, duplicate relative paths, etc.

    Raises:
        ValueError: If repo version is not valid.
    """
    validate_duplicate_content(version)
    validate_version_paths(version)



# Generated by Django 2.2.16 on 2020-09-08 16:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0047_improve_orphan_cleanup'),
    ]

    operations = [
        migrations.AlterField(
            model_name='artifact',
            name='md5',
            field=models.CharField(db_index=True, max_length=32, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha1',
            field=models.CharField(db_index=True, max_length=40, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha224',
            field=models.CharField(db_index=True, max_length=56, null=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha384',
            field=models.CharField(db_index=True, max_length=96, null=True, unique=True),
        ),
        migrations.AlterField(
            model_name='artifact',
            name='sha512',
            field=models.CharField(db_index=True, max_length=128, null=True, unique=True),
        ),
    ]



import os
import re
from gettext import gettext as _

from rest_framework import serializers
from rest_framework.validators import UniqueValidator

from pulpcore.app import models, settings
from pulpcore.app.serializers import (
    DetailIdentityField,
    DetailRelatedField,
    ExportIdentityField,
    ExportRelatedField,
    ModelSerializer,
    RelatedField,
    RelatedResourceField,
    RepositoryVersionRelatedField,
)
from pulpcore.constants import FS_EXPORT_CHOICES, FS_EXPORT_METHODS


class ExporterSerializer(ModelSerializer):
    """
    Base serializer for Exporters.
    """

    pulp_href = DetailIdentityField(view_name_pattern=r"exporter(-.*/.*)-detail")
    name = serializers.CharField(
        help_text=_("Unique name of the file system exporter."),
        validators=[UniqueValidator(queryset=models.Exporter.objects.all())],
    )

    @staticmethod
    def validate_path(value, check_is_dir=False):
        """
        Check if path is in ALLOWED_EXPORT_PATHS.

        Args:
            value: The user-provided value path to be validated.
            check_is_dir: If true, insure path ends with a directory

        Raises:
            ValidationError: When path is not in the ALLOWED_EXPORT_PATHS setting.

        Returns:
            The validated value.
        """
        for allowed_path in settings.ALLOWED_EXPORT_PATHS:
            user_provided_realpath = os.path.realpath(value)
            if user_provided_realpath.startswith(allowed_path):
                if check_is_dir:  # fail if exists and not-directory
                    if os.path.exists(user_provided_realpath) and not os.path.isdir(
                        user_provided_realpath
                    ):
                        raise serializers.ValidationError(
                            _("Path '{}' must be a directory path").format(value)
                        )
                return value
        raise serializers.ValidationError(
            _("Path '{}' is not an allowed export path").format(value)
        )

    class Meta:
        model = models.Exporter
        fields = ModelSerializer.Meta.fields + ("name",)


class ExportedResourceField(RelatedResourceField):
    class Meta:
        model = models.ExportedResource
        fields = []


class ExportSerializer(ModelSerializer):
    """
    Base serializer for Exports.
    """

    pulp_href = ExportIdentityField()

    task = RelatedField(
        help_text=_("A URI of the task that ran the Export."),
        queryset=models.Task.objects.all(),
        view_name="tasks-detail",
        required=False,
        allow_null=True,
    )

    exported_resources = ExportedResourceField(
        help_text=_("Resources that were exported."),
        many=True,
        read_only=True,
        view_name="None",  # This is a polymorphic field. The serializer does not need a view name.
    )

    params = serializers.JSONField(
        help_text=_("Any additional parameters that were used to create the export."),
        read_only=True,
    )

    class Meta:
        model = models.Export
        fields = ModelSerializer.Meta.fields + ("task", "exported_resources", "params")


class PulpExportSerializer(ExportSerializer):
    """
    Serializer for PulpExports.
    """

    output_file_info = serializers.JSONField(
        help_text=_("Dictionary of filename: sha256hash entries for export-output-file(s)"),
        read_only=True,
    )

    toc_info = serializers.JSONField(
        help_text=_("Filename and sha256-checksum of table-of-contents for this export"),
        read_only=True,
    )

    dry_run = serializers.BooleanField(
        help_text=_("Generate report on what would be exported and disk-space required."),
        default=False,
        required=False,
        write_only=True,
    )
    full = serializers.BooleanField(
        help_text=_("Do a Full (true) or Incremental (false) export."),
        default=True,
        required=False,
        write_only=True,
    )
    versions = RepositoryVersionRelatedField(
        help_text=_("List of explicit repo-version hrefs to export (replaces current_version)."),
        many=True,
        required=False,
        write_only=True,
    )

    MAX_CHUNK_BYTES = 1024 * 1024 * 1024 * 1024  # 1 TB
    chunk_size = serializers.CharField(
        help_text=_(
            "Chunk export-tarfile into pieces of chunk_size bytes. "
            + "Recognizes units of B/KB/MB/GB/TB. A chunk has a maximum "
            + "size of 1TB."
        ),
        required=False,
        write_only=True,
    )

    start_versions = RepositoryVersionRelatedField(
        help_text=_("List of explicit last-exported-repo-version hrefs (replaces last_export)."),
        many=True,
        required=False,
        write_only=True,
    )

    def _validate_versions_to_repos(self, the_versions):
        """
        If specifying repo-versions explicitly, must provide a version for each exporter-repository
        """
        # make sure counts match
        the_exporter = self.context.get("exporter", None)
        num_repos = the_exporter.repositories.count()
        if num_repos != len(the_versions):
            raise serializers.ValidationError(
                _(
                    "Number of versions ({}) does not match the number of Repositories ({}) for "
                    + "the owning  Exporter!"
                ).format(num_repos, len(the_versions))
            )

        # make sure the specified versions 'belong to' the exporter.repositories
        exporter_repos = set(the_exporter.repositories.all())
        version_repos = set([vers.repository for vers in the_versions])
        if exporter_repos != version_repos:
            raise serializers.ValidationError(
                _(
                    "Requested RepositoryVersions must belong to the Repositories named by the "
                    + "Exporter!"
                )
            )
        return the_versions

    def validate_versions(self, versions):
        return self._validate_versions_to_repos(versions)

    def validate_start_versions(self, start_versions):
        return self._validate_versions_to_repos(start_versions)

    def validate(self, data):
        # If we requested start_versions, make sure we did not forget to specify full=False
        if data.get("start_versions", None) and data.get("full", True):
            raise serializers.ValidationError(
                _("start_versions is only valid for incremental exports (full=False)")
            )

        # If we requested full=False, make sure we either specified start_versions= or
        # have a previous-export for our Exporter.
        if not data.get("full", True):
            the_exporter = self.context.get("exporter", None)
            if not data.get("start_versions", None) and not the_exporter.last_export:
                raise serializers.ValidationError(
                    _(
                        "Incremental export can only be requested when there is a previous export "
                        + "or start_versions= has been specified."
                    )
                )
        return super().validate(data)

    @staticmethod
    def _parse_size(size):
        try:
            # based on https://stackoverflow.com/a/42865957/2002471
            units = {"B": 1, "KB": 2**10, "MB": 2**20, "GB": 2**30, "TB": 2**40}
            size = size.upper()
            if not re.match(r" ", size):
                size = re.sub(r"([KMGT]?B)", r" \1", size)
            number, unit = [string.strip() for string in size.split()]
            return int(float(number) * units[unit])
        except ValueError:
            raise serializers.ValidationError(
                _("chunk_size '{}' is not valid (valid units are B/KB/MB/GB/TB)").format(size)
            )

    def validate_chunk_size(self, chunk_size):
        the_size = self._parse_size(chunk_size)
        if the_size <= 0:
            raise serializers.ValidationError(
                _("Chunk size {} is not greater than zero!").format(the_size)
            )
        if the_size > self.MAX_CHUNK_BYTES:
            raise serializers.ValidationError(
                _("Chunk size in bytes {} is greater than max-chunk-size {}!").format(
                    the_size, self.MAX_CHUNK_BYTES
                )
            )
        return the_size

    class Meta:
        model = models.PulpExport
        fields = ExportSerializer.Meta.fields + (
            "full",
            "dry_run",
            "versions",
            "chunk_size",
            "output_file_info",
            "start_versions",
            "toc_info",
        )


class PulpExporterSerializer(ExporterSerializer):
    """
    Serializer for pulp exporters.
    """

    path = serializers.CharField(help_text=_("File system directory to store exported tar.gzs."))

    repositories = DetailRelatedField(
        view_name_pattern=r"repositories(-.*/.*)-detail",
        queryset=models.Repository.objects.all(),
        many=True,
    )
    last_export = ExportRelatedField(
        help_text=_("Last attempted export for this PulpExporter"),
        queryset=models.PulpExport.objects.all(),
        many=False,
        required=False,
        allow_null=True,
    )

    class Meta:
        model = models.PulpExporter
        fields = ExporterSerializer.Meta.fields + ("path", "repositories", "last_export")


class FilesystemExportSerializer(ExportSerializer):
    """
    Serializer for FilesystemExports.
    """

    publication = DetailRelatedField(
        required=False,
        help_text=_("A URI of the publication to be exported."),
        view_name_pattern=r"publications(-.*/.*)-detail",
        queryset=models.Publication.objects.all(),
        write_only=True,
    )

    repository_version = RepositoryVersionRelatedField(
        help_text=_("A URI of the repository version export."),
        required=False,
        write_only=True,
    )

    def validate(self, data):
        if ("publication" not in data and "repository_version" not in data) or (
            "publication" in data and "repository_version" in data
        ):
            raise serializers.ValidationError(
                _("publication or repository_version must either be supplied but not both.")
            )
        return data

    class Meta:
        model = models.FilesystemExport
        fields = ExportSerializer.Meta.fields + ("publication", "repository_version")


class FilesystemExporterSerializer(ExporterSerializer):
    """
    Serializer for FilesystemExporters.
    """

    path = serializers.CharField(help_text=_("File system location to export to."))

    method = serializers.ChoiceField(
        help_text=_("Method of exporting"),
        choices=(FS_EXPORT_CHOICES),
        default=FS_EXPORT_METHODS.WRITE,
    )

    class Meta:
        model = models.FilesystemExporter
        fields = ExporterSerializer.Meta.fields + ("path", "method")



# Generated by Django 2.2.15 on 2020-09-09 19:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0046_task__resource_job_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contentartifact',
            name='artifact',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='content_memberships', to='core.Artifact'),
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='content',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='contentguard',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='exporter',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='importer',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='publication',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='remote',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
        migrations.AlterField(
            model_name='repository',
            name='pulp_type',
            field=models.TextField(db_index=True, default=None),
        ),
    ]



"""Utilities for pulpcore API tests that require the use of a plugin."""
from functools import partial
from unittest import SkipTest

from pulp_smash import api, cli, config, selectors, utils
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import gen_remote, gen_repo, require_pulp_3, require_pulp_plugins, sync

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_PATH,
    FILE_FIXTURE_MANIFEST_URL,
    FILE_PUBLICATION_PATH,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ExportersPulpApi,
)
from pulpcore.client.pulp_file import DistributionsFileApi
from pulpcore.client.pulp_file import ApiClient as FileApiClient

skip_if = partial(selectors.skip_if, exc=SkipTest)


def set_up_module():
    """Conditions to skip tests.

    Skip tests if not testing Pulp 3, or if either pulpcore or pulp_file
    aren't installed.
    """
    require_pulp_3(SkipTest)
    require_pulp_plugins({"core", "file"}, SkipTest)


def populate_pulp(cfg, url=None):
    """Add file contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp
        application.
    :param url: The URL to a file repository's ``PULP_MANIFEST`` file. Defaults
        to :data:`pulp_smash.constants.FILE_FIXTURE_URL` + ``PULP_MANIFEST``.
    :returns: A list of dicts, where each dict describes one file content in
        Pulp.
    """
    if url is None:
        url = FILE_FIXTURE_MANIFEST_URL

    client = api.Client(cfg, api.page_handler)
    remote = {}
    repo = {}
    try:
        remote.update(client.post(FILE_REMOTE_PATH, gen_remote(url)))
        repo.update(client.post(FILE_REPO_PATH, gen_repo()))
        sync(cfg, remote, repo)
    finally:
        if remote:
            client.delete(remote["pulp_href"])
        if repo:
            client.delete(repo["pulp_href"])
    return client.get(FILE_CONTENT_PATH)


def gen_file_client():
    """Return an OBJECT for file client."""
    configuration = config.get_config().get_bindings_config()
    return FileApiClient(configuration)


def gen_file_remote(url=FILE_FIXTURE_MANIFEST_URL, **kwargs):
    """Return a semi-random dict for use in creating a file Remote.

    :param url: The URL of an external content source.
    """
    return gen_remote(url, **kwargs)


def create_file_publication(cfg, repo, version_href=None):
    """Create a file publication.

    :param pulp_smash.config.PulpSmashConfig cfg: Information about the Pulp
        host.
    :param repo: A dict of information about the repository.
    :param version_href: A href for the repo version to be published.
    :returns: A publication. A dict of information about the just created
        publication.
    """
    if version_href:
        body = {"repository_version": version_href}
    else:
        body = {"repository": repo["pulp_href"]}
    return api.Client(cfg).post(FILE_PUBLICATION_PATH, body)


def create_repo_and_versions(syncd_repo, repo_api, versions_api, content_api):
    """Create a repo with multiple versions.

    :param syncd_repo: A Repository that has at least three Content-units for us to copy from.
    :param pulpcore.client.pulp_file.RepositoriesFileApi repo_api: client to talk to the Repository
        API
    :param pulpcore.client.pulp_file.RepositoriesFileVersionsApi versions_api: client to talk to
        the RepositoryVersions API
    :param pulpcore.client.pulp_file.ContentFilesApi content_api: client to talk to the Content API
    :returns: A (FileRepository, [FileRepositoryVersion...]) tuple
    """
    # Create a new file-repo
    a_repo = repo_api.create(gen_repo())
    # get a list of all the files from one of our existing repos
    file_list = content_api.list(repository_version=syncd_repo.latest_version_href)
    # copy files from repositories[0] into new, one file at a time
    results = file_list.results
    for a_file in results:
        href = a_file.pulp_href
        modify_response = repo_api.modify(a_repo.pulp_href, {"add_content_units": [href]})
        monitor_task(modify_response.task)
    # get all versions of that repo
    versions = versions_api.list(a_repo.pulp_href, ordering="number")
    return a_repo, versions


def delete_exporter(exporter):
    """
    Utility routine to delete an exporter and any exported files
    :param exporter : PulpExporter to delete
    """
    cfg = config.get_config()
    cli_client = cli.Client(cfg)
    core_client = CoreApiClient(configuration=cfg.get_bindings_config())
    exporter_api = ExportersPulpApi(core_client)
    cmd = ("rm", "-rf", exporter.path)

    cli_client.run(cmd, sudo=True)
    result = exporter_api.delete(exporter.pulp_href)
    monitor_task(result.task)


def create_distribution(repository_href=None):
    """Utility to create a pulp_file distribution."""
    file_client = gen_file_client()
    distro_api = DistributionsFileApi(file_client)

    body = {"name": utils.uuid4(), "base_path": utils.uuid4()}
    if repository_href:
        body["repository"] = repository_href

    result = distro_api.create(body)
    distro_href = monitor_task(result.task).created_resources[0]
    distro = distro_api.read(distro_href)
    return distro


CREATE_USER_CMD = [
    "from django.contrib.auth import get_user_model",
    "from django.urls import resolve",
    "from guardian.shortcuts import assign_perm",
    "",
    "user = get_user_model().objects.create(username='{username}')",
    "user.set_password('{password}')",
    "user.save()",
    "",
    "for permission in {model_permissions!r}:",
    "    assign_perm(permission, user)",
    "",
    "for permission, obj_url in {object_permissions!r}:",
    "    func, _, kwargs = resolve(obj_url)",
    "    obj = func.cls.queryset.get(pk=kwargs['pk'])",
    "    assign_perm(permission, user, obj)",
]


DELETE_USER_CMD = [
    "from django.contrib.auth import get_user_model",
    "get_user_model().objects.get(username='{username}').delete()",
]


def gen_user(cfg=config.get_config(), model_permissions=None, object_permissions=None):
    """Create a user with a set of permissions in the pulp database."""
    cli_client = cli.Client(cfg)

    if model_permissions is None:
        model_permissions = []

    if object_permissions is None:
        object_permissions = []

    user = {
        "username": utils.uuid4(),
        "password": utils.uuid4(),
        "model_permissions": model_permissions,
        "object_permissions": object_permissions,
    }
    utils.execute_pulpcore_python(
        cli_client,
        "\n".join(CREATE_USER_CMD).format(**user),
    )

    api_config = cfg.get_bindings_config()
    api_config.username = user["username"]
    api_config.password = user["password"]
    user["core_api_client"] = CoreApiClient(api_config)
    user["api_client"] = FileApiClient(api_config)
    user["distribution_api"] = DistributionsFileApi(user["api_client"])
    return user


def del_user(user, cfg=config.get_config()):
    """Delete a user from the pulp database."""
    cli_client = cli.Client(cfg)
    utils.execute_pulpcore_python(
        cli_client,
        "\n".join(DELETE_USER_CMD).format(**user),
    )



import logging
import shutil
from gettext import gettext as _

from django.conf import settings
from django.core.files.storage import default_storage
from drf_spectacular.utils import extend_schema
from rest_framework.response import Response
from rest_framework.views import APIView

from pulpcore.app.apps import pulp_plugin_configs
from pulpcore.app.models.status import ContentAppStatus
from pulpcore.app.models.task import Worker
from pulpcore.app.serializers.status import StatusSerializer
from pulpcore.app.redis_connection import get_redis_connection

_logger = logging.getLogger(__name__)


def _disk_usage():
    if settings.DEFAULT_FILE_STORAGE == "pulpcore.app.models.storage.FileSystem":
        try:
            return shutil.disk_usage(default_storage.location)
        except Exception:
            _logger.exception(_("Failed to determine disk usage"))

    return None


class StatusView(APIView):
    """
    Returns status information about the application
    """

    # allow anyone to access the status api
    authentication_classes = []
    permission_classes = []

    @extend_schema(
        summary="Inspect status of Pulp",
        operation_id="status_read",
        responses={200: StatusSerializer},
    )
    def get(self, request):
        """
        Returns status and app information about Pulp.

        Information includes:
         * version of pulpcore and loaded pulp plugins
         * known workers
         * known content apps
         * database connection status
         * redis connection status
         * disk usage information
        """
        versions = []
        for app in pulp_plugin_configs():
            versions.append({"component": app.label, "version": app.version})

        if settings.CACHE_ENABLED:
            redis_status = {"connected": self._get_redis_conn_status()}
        else:
            redis_status = None

        db_status = {"connected": self._get_db_conn_status()}

        try:
            online_workers = Worker.objects.online_workers()
        except Exception:
            online_workers = None

        try:
            online_content_apps = ContentAppStatus.objects.online()
        except Exception:
            online_content_apps = None

        data = {
            "versions": versions,
            "online_workers": online_workers,
            "online_content_apps": online_content_apps,
            "database_connection": db_status,
            "redis_connection": redis_status,
            "storage": _disk_usage(),
        }

        context = {"request": request}
        serializer = StatusSerializer(data, context=context)
        return Response(serializer.data)

    @staticmethod
    def _get_db_conn_status():
        """
        Returns True if pulp is connected to the database

        Returns:
            bool: True if there's a db connection. False otherwise.
        """
        try:
            Worker.objects.count()
        except Exception:
            _logger.exception(_("Cannot connect to database during status check."))
            return False
        else:
            return True

    @staticmethod
    def _get_redis_conn_status():
        """
        Returns True if pulp can connect to Redis

        Returns:
            bool: True if pulp can connect to Redis. False otherwise.
        """
        conn = get_redis_connection()
        try:
            conn.ping()
        except Exception:
            _logger.error(_("Connection to Redis failed during status check!"))
            return False
        else:
            return True



# Models are exposed selectively in the versioned plugin API.
# Any models defined in the pulpcore.plugin namespace should probably be proxy models.

from pulpcore.app.models import (  # noqa
    AlternateContentSource,
    AlternateContentSourcePath,
    AccessPolicy,
    AutoAddObjPermsMixin,
    AutoDeleteObjPermsMixin,
    Artifact,
    AsciiArmoredDetachedSigningService,
    BaseModel,
    Content,
    ContentArtifact,
    ContentManager,
    ContentGuard,
    CreatedResource,
    Distribution,
    Export,
    Exporter,
    GroupProgressReport,
    Import,
    Importer,
    FilesystemExporter,
    Label,
    MasterModel,
    ProgressReport,
    Publication,
    PublishedArtifact,
    PublishedMetadata,
    PulpTemporaryFile,
    Repository,
    Remote,
    RemoteArtifact,
    RepositoryContent,
    RepositoryVersion,
    SigningService,
    Task,
    TaskGroup,
    Upload,
    UploadChunk,
)



"""Tests that perform actions over publications."""
import unittest
from itertools import permutations

from pulp_smash import api, config
from pulp_smash.pulp3.utils import gen_distribution, gen_repo, get_content, modify_repo, sync
from requests.exceptions import HTTPError

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_NAME,
    FILE_DISTRIBUTION_PATH,
    FILE_PUBLICATION_PATH,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    create_file_publication,
    gen_file_remote,
)
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa
from pulpcore.tests.functional.api.using_plugin.utils import skip_if
from pulpcore.tests.functional.api.utils import parse_date_from_string


class PublicationsTestCase(unittest.TestCase):
    """Perform actions over publications."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.page_handler)
        cls.client_echo = api.Client(cls.cfg, api.echo_handler)
        cls.remote = {}
        cls.publication = {}
        cls.repo = {}
        try:
            cls.repo.update(cls.client.post(FILE_REPO_PATH, gen_repo()))
            cls.repo_initial_version = cls.repo["latest_version_href"]
            body = gen_file_remote()
            cls.remote.update(cls.client.post(FILE_REMOTE_PATH, body))
            sync(cls.cfg, cls.remote, cls.repo)
            # update to get latest_version_href
            cls.repo.update(cls.client.get(cls.repo["pulp_href"]))
        except Exception:
            cls.tearDownClass()
            raise

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variables."""
        for resource in (cls.remote, cls.repo):
            if resource:
                cls.client.delete(resource["pulp_href"])

    def test_01_create_file_publication(self):
        """Create a publication."""
        self.publication.update(create_file_publication(self.cfg, self.repo))

    @skip_if(bool, "publication", False)
    def test_02_read_publication(self):
        """Read a publication by its href."""
        publication = self.client.get(self.publication["pulp_href"])
        for key, val in self.publication.items():
            with self.subTest(key=key):
                self.assertEqual(publication[key], val)

    @skip_if(bool, "publication", False)
    def test_02_read_publication_with_specific_fields(self):
        """Read a publication by its href providing specific field list.

        Permutate field list to ensure different combinations on result.
        """
        fields = ("pulp_href", "pulp_created", "distributions")
        for field_pair in permutations(fields, 2):
            # ex: field_pair = ('pulp_href', 'pulp_created)
            with self.subTest(field_pair=field_pair):
                publication = self.client.get(
                    self.publication["pulp_href"], params={"fields": ",".join(field_pair)}
                )
                self.assertEqual(sorted(field_pair), sorted(publication.keys()))

    @skip_if(bool, "publication", False)
    def test_02_read_publication_without_specific_fields(self):
        """Read a publication by its href excluding specific fields."""
        # requests doesn't allow the use of != in parameters.
        url = "{}?exclude_fields=distributions".format(self.publication["pulp_href"])
        publication = self.client.get(url)
        self.assertNotIn("distributions", publication.keys())

    @skip_if(bool, "publication", False)
    def test_02_read_publications_filter_repo_version(self):
        """Read a publication by its repository version."""
        publications = self.client.get(
            FILE_PUBLICATION_PATH, params={"repository_version": self.repo["latest_version_href"]}
        )
        self.assertEqual(len(publications), 1, publications)
        for key, val in self.publication.items():
            with self.subTest(key=key):
                self.assertEqual(publications[0][key], val)

    @skip_if(bool, "publication", False)
    def test_02_read_publications_filter_repo_version_no_match(self):
        """Filter by repo version for which no publication exists."""
        publications = self.client.get(
            FILE_PUBLICATION_PATH, params={"repository_version": self.repo_initial_version}
        )
        self.assertFalse(publications)

    @skip_if(bool, "publication", False)
    def test_02_read_publications_filter_repo_version_invalid(self):
        """Filter by a repo version that does not exist."""
        invalid_repo_version = self.repo["versions_href"] + "123456789/"
        response = self.client_echo.get(
            FILE_PUBLICATION_PATH, params={"repository_version": invalid_repo_version}
        )
        self.assertEqual(response.status_code, 400)
        self.assertIn("not found for repositoryversion", response.text)

    @skip_if(bool, "publication", False)
    def test_02_read_publications_filter_created_time(self):
        """Read a publication by its created time."""
        publications = self.client.get(
            FILE_PUBLICATION_PATH, params={"pulp_created": self.publication["pulp_created"]}
        )
        self.assertEqual(len(publications), 1, publications)
        for key, val in self.publication.items():
            with self.subTest(key=key):
                self.assertEqual(publications[0][key], val)

    @skip_if(bool, "publication", False)
    def test_02_read_publications_filter_created_time_no_match(self):
        """Filter for created time for which no publication exists."""
        publications = self.client.get(
            FILE_PUBLICATION_PATH, params={"pulp_created": self.repo["pulp_created"]}
        )
        self.assertFalse(publications)

    @skip_if(bool, "publication", False)
    @unittest.skip("distribution filter not implemented")
    def test_02_read_publications_filter_distribution(self):
        """Read a publication by its distribution."""
        body = gen_distribution()
        body["publication"] = self.publication["pulp_href"]
        distribution = self.client.using_handler(api.task_handler).post(
            FILE_DISTRIBUTION_PATH, body
        )
        self.addCleanup(self.client.delete, distribution["pulp_href"])

        self.publication.update(self.client.get(self.publication["pulp_href"]))
        publications = self.client.get(
            FILE_PUBLICATION_PATH, params={"distributions": distribution["pulp_href"]}
        )
        self.assertEqual(len(publications), 1, publications)
        for key, val in self.publication.items():
            with self.subTest(key=key):
                self.assertEqual(publications[0][key], val)

    @skip_if(bool, "publication", False)
    def test_06_publication_create_order(self):
        """Assert that publications are ordered by created time.

        This test targets the following issues:

        * `Pulp Smash #954 <https://github.com/pulp/pulp-smash/issues/954>`_
        * `Pulp #3576 <https://pulp.plan.io/issues/3576>`_
        """
        # Create more 2 publications for the same repo
        for _ in range(2):
            create_file_publication(self.cfg, self.repo)

        # Read publications
        publications = self.client.get(FILE_PUBLICATION_PATH)
        self.assertEqual(len(publications), 3)

        # Assert publications are ordered by pulp_created field in descending order
        for i, publication in enumerate(publications[:-1]):
            self.assertGreater(
                parse_date_from_string(publication["pulp_created"]),  # Current
                parse_date_from_string(publications[i + 1]["pulp_created"]),  # Prev
            )

    @skip_if(bool, "publication", False)
    def test_07_delete(self):
        """Delete a publication."""
        self.client.delete(self.publication["pulp_href"])
        with self.assertRaises(HTTPError):
            self.client.get(self.publication["pulp_href"])


class PublicationRepositoryParametersTestCase(unittest.TestCase):
    """Explore publication creation using repository and repository version.

    This test targets the following issue:

    * `Pulp #4854 <https://pulp.plan.io/issues/4854>`_
    * `Pulp #4874 <https://pulp.plan.io/issues/4874>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)

    def test_create_only_using_repoversion(self):
        """Create a publication only using repository version."""
        repo = self.create_sync_repo()
        for file_content in get_content(repo)[FILE_CONTENT_NAME]:
            modify_repo(self.cfg, repo, remove_units=[file_content])
        version_href = self.client.get(repo["versions_href"])[1]["pulp_href"]
        publication = create_file_publication(self.cfg, repo, version_href)
        self.addCleanup(self.client.delete, publication["pulp_href"])

        self.assertEqual(publication["repository_version"], version_href, publication)

    def test_create_repo_repoversion(self):
        """Create a publication using repository and repository version."""
        repo = self.create_sync_repo()
        version_href = self.client.get(repo["versions_href"])[0]["pulp_href"]

        with self.assertRaises(HTTPError) as ctx:
            self.client.using_handler(api.json_handler).post(
                FILE_PUBLICATION_PATH,
                {"repository_version": version_href, "repository": repo["pulp_href"]},
            )

        for key in ("repository", "repository_version", "not", "both"):
            self.assertIn(
                key,
                ctx.exception.response.json()["non_field_errors"][0].lower(),
                ctx.exception.response,
            )

    def create_sync_repo(self):
        """Create and sync a repository.

        Given the number of times to be synced.
        """
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])

        remote = self.client.post(FILE_REMOTE_PATH, gen_file_remote())
        self.addCleanup(self.client.delete, remote["pulp_href"])

        sync(self.cfg, remote, repo)
        return self.client.get(repo["pulp_href"])



# Generated by Django 2.2.23 on 2021-06-03 20:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0070_rename_retained_versions'),
    ]

    operations = [
        migrations.CreateModel(
            name='FilesystemExport',
            fields=[
                ('export_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Export')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.export',),
        ),
        migrations.CreateModel(
            name='FilesystemExporter',
            fields=[
                ('exporter_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='core_fs_exporter', serialize=False, to='core.Exporter')),
                ('path', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_fs_exporter',
            },
            bases=('core.exporter',),
        ),
    ]



from django.db import models

from pulpcore.app.models import (
    BaseModel,
    MasterModel,
)
from .repository import Repository


class Import(BaseModel):
    """
    A model that represents imports into Pulp.

    Fields:
        params (models.JSONField): A set of parameters used to run the import

    Relations:
        task (models.ForeignKey): The Task that ran the import
        importer (models.ForeignKey): The Importer that imported the export
    """

    params = models.JSONField(null=True)
    task = models.ForeignKey("Task", on_delete=models.PROTECT)
    importer = models.ForeignKey("Importer", on_delete=models.CASCADE)


class Importer(MasterModel):
    """
    A base model that provides logic to import data into Pulp.

    Can be extended by plugins to provide import functionality.

    Fields:
        name (models.TextField): The importer unique name.
    """

    name = models.TextField(db_index=True, unique=True)


class PulpImporter(Importer):
    """
    A model that can be used to import exports from other Pulp instances.
    """

    TYPE = "pulp"

    @property
    def repo_mapping(self):
        return {repo.source_repo: repo.repository.name for repo in self.repo_map.all()}

    @repo_mapping.setter
    def repo_mapping(self, mapping):
        self.repo_map.all().delete()
        for source, repo_name in mapping.items():
            repo = Repository.objects.get(name=repo_name)
            self.repo_map.create(source_repo=source, repository=repo)

    class Meta:
        default_related_name = "%(app_label)s_pulp_importer"


class PulpImporterRepository(BaseModel):
    """
    A model that maps repo names in an export to repos in Pulp.

    Fields:
        source_repo (models.TextField): The name of the repo in the export

    Relations:
        pulp_importer (models.ForeignKey): The associated Pulp importer
        repository (models.ForeignKey): The repository in Pulp
    """

    source_repo = models.TextField()
    pulp_importer = models.ForeignKey(
        PulpImporter, related_name="repo_map", on_delete=models.CASCADE
    )
    repository = models.ForeignKey(Repository, on_delete=models.CASCADE)


class PulpImport(Import):
    """A model that represents imports into Pulp from another Pulp instance."""

    class Meta:
        default_related_name = "%(app_label)s_pulp_export"



import asyncio
import tempfile

from .api import create_pipeline, EndStage
from .artifact_stages import (
    ACSArtifactHandler,
    ArtifactDownloader,
    ArtifactSaver,
    QueryExistingArtifacts,
    RemoteArtifactSaver,
)
from .content_stages import (
    ContentAssociation,
    ContentSaver,
    QueryExistingContents,
    ResolveContentFutures,
)


class DeclarativeVersion:
    def __init__(self, first_stage, repository, mirror=False, acs=False):
        """
        A pipeline that creates a new :class:`~pulpcore.plugin.models.RepositoryVersion` from a
        stream of :class:`~pulpcore.plugin.stages.DeclarativeContent` objects.

        The plugin writer needs to specify a first_stage that will create a
        :class:`~pulpcore.plugin.stages.DeclarativeContent` object for each Content unit that should
        exist in the :class:`~pulpcore.plugin.models.RepositoryVersion`.

        The pipeline stages perform the following steps by default:

        1. Create the new :class:`~pulpcore.plugin.models.RepositoryVersion`
        2. Use the provided `first_stage` to construct
           :class:`~pulpcore.plugin.stages.DeclarativeContent`
        3. Query existing artifacts to determine which are already local to Pulp with
           :class:`~pulpcore.plugin.stages.QueryExistingArtifacts`
        4. Download any undownloaded :class:`~pulpcore.plugin.models.Artifact` objects with
           :class:`~pulpcore.plugin.stages.ArtifactDownloader`
        5. Save the newly downloaded :class:`~pulpcore.plugin.models.Artifact` objects with
           :class:`~pulpcore.plugin.stages.ArtifactSaver`
        6. Query for Content units already present in Pulp with
           :class:`~pulpcore.plugin.stages.QueryExistingContents`
        7. Save new Content units not yet present in Pulp with
           :class:`~pulpcore.plugin.stages.ContentSaver`
        8. Attach :class:`~pulpcore.plugin.models.RemoteArtifact` to the
           :class:`~pulpcore.plugin.models.Content` via
           :class:`~pulpcore.plugin.stages.RemoteArtifactSaver`
        9. Resolve the attached :class:`~asyncio.Future` of
           :class:`~pulpcore.plugin.stages.DeclarativeContent` with
           :class:`~pulpcore.plugin.stages.ResolveContentFutures`
        10. Associate all content units with the new
            :class:`~pulpcore.plugin.models.RepositoryVersion` with
            :class:`~pulpcore.plugin.stages.ContentAssociation`
        11. Unassociate any content units not declared in the stream (only when mirror=True)
            with :class:`~pulpcore.plugin.stages.ContentUnassociation`

        To do this, the plugin writer should subclass the
        :class:`~pulpcore.plugin.stages.Stage` class and define its
        :meth:`run()` interface which returns a coroutine. This coroutine should
        download metadata, create the corresponding
        :class:`~pulpcore.plugin.stages.DeclarativeContent` objects, and put them into the
        :class:`asyncio.Queue` via :meth:`put()` to send them down the pipeline. For example::

            class MyFirstStage(Stage):

                def __init__(remote):
                    self.remote = remote

                async def run(self):
                    downloader = remote.get_downloader(url=remote.url)
                    result = await downloader.run()
                    for entry in read_my_metadata_file_somehow(result.path)
                        unit = MyContent(entry)  # make the content unit in memory-only
                        artifact = Artifact(entry)  # make Artifact in memory-only
                        da = DeclarativeArtifact(artifact, url, entry.relative_path, self.remote)
                        dc = DeclarativeContent(content=unit, d_artifacts=[da])
                        await self.put(dc)

        To use your first stage with the pipeline you have to instantiate the subclass and pass it
        to :class:`~pulpcore.plugin.stages.DeclarativeVersion`.

        1. Create the instance of the subclassed :class:`~pulpcore.plugin.stages.Stage` object.
        2. Create the :class:`~pulpcore.plugin.stages.DeclarativeVersion` instance, passing the
           :class:`~pulpcore.plugin.stages.Stage` subclass instance to it
        3. Call the :meth:`~pulpcore.plugin.stages.DeclarativeVersion.create` method on your
           :class:`~pulpcore.plugin.stages.DeclarativeVersion` instance

        Here is an example::

            first_stage = MyFirstStage(remote)
            DeclarativeVersion(first_stage, repository_version).create()

        Args:
            first_stage (:class:`~pulpcore.plugin.stages.Stage`): The first stage to receive
                :class:`~pulpcore.plugin.stages.DeclarativeContent` from.
            repository (:class:`~pulpcore.plugin.models.Repository`): The repository receiving the
                new version.
            mirror (bool): 'True' removes content units from the
                :class:`~pulpcore.plugin.models.RepositoryVersion` that are not
                requested in the :class:`~pulpcore.plugin.stages.DeclarativeVersion` stream.
                'False' (additive) only adds content units observed in the
                :class:`~pulpcore.plugin.stages.DeclarativeVersion stream`, and does not remove any
                pre-existing units in the :class:`~pulpcore.plugin.models.RepositoryVersion`.
                'False' is the default.
            acs (bool): When set to 'True' a new stage is added to look for
                Alternate Content Sources.

        """
        self.first_stage = first_stage
        self.repository = repository
        self.mirror = mirror
        self.acs = acs

    def pipeline_stages(self, new_version):
        """
        Build the list of pipeline stages feeding into the ContentAssociation stage.

        Plugin-writers may override this method to build a custom pipeline. This
        can be achieved by returning a list with different stages or by extending
        the list returned by this method.

        Args:
            new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`): The
                new repository version that is going to be built.

        Returns:
            list: List of :class:`~pulpcore.plugin.stages.Stage` instances

        """
        pipeline = [
            self.first_stage,
            QueryExistingArtifacts(),
        ]
        if self.acs:
            pipeline.append(ACSArtifactHandler())
        pipeline.extend(
            [
                ArtifactDownloader(),
                ArtifactSaver(),
                QueryExistingContents(),
                ContentSaver(),
                RemoteArtifactSaver(),
                ResolveContentFutures(),
            ]
        )
        return pipeline

    def create(self):
        """
        Perform the work. This is the long-blocking call where all syncing occurs.

        Returns: The created RepositoryVersion or None if it represents no change from the latest.
        """
        with tempfile.TemporaryDirectory(dir="."):
            with self.repository.new_version() as new_version:
                loop = asyncio.get_event_loop()
                stages = self.pipeline_stages(new_version)
                stages.append(ContentAssociation(new_version, self.mirror))
                stages.append(EndStage())
                pipeline = create_pipeline(stages)
                loop.run_until_complete(pipeline)

        return new_version if new_version.complete else None



from unittest import TestCase, mock

from pulpcore.app import models, util


class TestViewNameForModel(TestCase):
    def test_repository(self):
        """
        Use Repository as an example that should work.
        """
        ret = util.get_view_name_for_model(models.Artifact(), "foo")
        self.assertEqual(ret, "artifacts-foo")

    @mock.patch.object(util, "get_viewset_for_model")
    def test_not_found(self, mock_viewset_for_model):
        """
        Given an unknown viewset (in this case a Mock()), this should raise LookupError.
        """
        self.assertRaises(LookupError, util.get_view_name_for_model, mock.Mock(), "foo")



from django.contrib.auth.backends import RemoteUserBackend
from rest_framework.authentication import RemoteUserAuthentication

from pulpcore.app import settings


class PulpRemoteUserAuthentication(RemoteUserAuthentication):

    header = settings.REMOTE_USER_ENVIRON_NAME


class PulpNoCreateRemoteUserBackend(RemoteUserBackend):

    create_unknown_user = False  # Configure RemoteUserBackend to not create users



import os

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from pkg_resources import get_distribution

from pulpcore.app.apps import pulp_plugin_configs
from pulpcore.app import models

# a little cache so viewset_for_model doesn't have iterate over every app every time
_model_viewset_cache = {}


# based on their name, viewset_for_model and view_name_for_model look like they should
# live over in the viewsets namespace, but these tools exist for serializers, which are
# depended on by viewsets. They're defined here because they're used here, and to avoid
# odd import dependencies.
def get_viewset_for_model(model_obj):
    """
    Given a Model instance or class, return the registered ViewSet for that Model
    """
    # model_obj can be an instance or class, force it to class
    model_class = model_obj._meta.model
    if model_class in _model_viewset_cache:
        return _model_viewset_cache[model_class]

    # cache miss, fill in the cache while we look for our matching viewset
    model_viewset = None
    # go through the viewset registry to find the viewset for the passed-in model
    for app in pulp_plugin_configs():
        for model, viewsets in app.named_viewsets.items():
            # There may be multiple viewsets for a model. In this
            # case, we can't reverse the mapping.
            if len(viewsets) == 1:
                viewset = viewsets[0]
                _model_viewset_cache.setdefault(model, viewset)
                if model is model_class:
                    model_viewset = viewset
                    break
        if model_viewset is not None:
            break

    if model_viewset is None:
        raise LookupError("Could not determine ViewSet base name for model {}".format(model_class))

    return viewset


def get_view_name_for_model(model_obj, view_action):
    """
    Given a Model instance or class, return the correct view name for that ViewSet view.

    This is the "glue" that generates view names dynamically based on a model object.

    Args:
        model_obj (pulpcore.app.models.Model): a Model that should have a ViewSet
        view_action (str): name of the view action as expected by DRF. See their docs for details.

    Returns:
        str: view name for the correct ViewSet

    Raises:
        LookupError: if no ViewSet is found for the Model
    """
    # Import this here to prevent out-of-order plugin discovery
    from pulpcore.app.urls import all_routers

    if isinstance(model_obj, models.MasterModel):
        model_obj = model_obj.cast()
    viewset = get_viewset_for_model(model_obj)

    # return the complete view name, joining the registered viewset base name with
    # the requested view method.
    for router in all_routers:
        for pattern, registered_viewset, base_name in router.registry:
            if registered_viewset is viewset:
                return "-".join((base_name, view_action))
    raise LookupError("view not found")


def batch_qs(qs, batch_size=1000):
    """
    Returns a queryset batch in the given queryset.

    Usage:
        # Make sure to order your querset
        article_qs = Article.objects.order_by('id')
        for qs in batch_qs(article_qs):
            for article in qs:
                print article.body
    """
    total = qs.count()
    for start in range(0, total, batch_size):
        end = min(start + batch_size, total)
        yield qs[start:end]


def get_version_from_model(in_model):
    """
    Return a tuple (dist-label, version) for the distribution that 'owns' the model

    Args:
        in_model (models.Model): model whose owning-plugin-version we need

    Returns:
        (str, str): tuple containing owning-plugin's (distribution, version)
    """
    app_label = ContentType.objects.get_for_model(in_model).app_label
    app_config_module = apps.get_app_config(app_label).name
    maybe_the_distribution_name = app_config_module.split(".")[0]
    version = get_distribution(maybe_the_distribution_name).version  # hope for the best!
    return maybe_the_distribution_name, version


def get_view_urlpattern(view):
    """
    Get a full urlpattern for a view which includes a parent urlpattern if it exists.
    E.g. for repository versions the urlpattern is just `versions` without its parent_viewset
    urlpattern.

    Args:
        view(subclass rest_framework.viewsets.GenericViewSet): The view being requested.

    Returns:
        str: a full urlpattern for a specified view/viewset
    """
    if hasattr(view, "parent_viewset") and view.parent_viewset:
        return os.path.join(view.parent_viewset.urlpattern(), view.urlpattern())
    return view.urlpattern()


def get_request_without_query_params(context):
    """
    Remove query parameters from a request object.

    Removing query parameters should not influence the features provided by the library
    'djangorestframework-queryfields'. But, once a user selects which fields should be sent in the
    API response, any other fields are going to be filtered out even though they are still required
    for serialization purposes in the future.

    Some serializers need to be aware of the context that triggered the operation. For example, when
    a Task is serialized, created resources need to be serialized as well. Missing context lead to
    invalid serialization if the created resources do not have knowledge about the initial request.

    This method modifies the request object in-place, meaning that all other objects that reference
    this object may be affected.

    Args:
        context (dict): A context containing the request object.

    Returns:
        rest_framework.request.Request: a request object without query parameters
    """
    request = context.get("request")

    if request is not None:
        request.query_params._mutable = True
        request.query_params.clear()
        request.query_params._mutable = False

    return request



"""
This module contains custom filters that might be used by more than one ViewSet.
"""
import re
from gettext import gettext as _
from urllib.parse import urlparse
from uuid import UUID

from django.urls import Resolver404, resolve
from django_filters import BaseInFilter, CharFilter, DateTimeFilter, Filter
from django_filters.fields import IsoDateTimeField
from rest_framework import serializers
from rest_framework.serializers import ValidationError as DRFValidationError

from pulpcore.app.models import ContentArtifact, Label, RepositoryVersion
from pulpcore.app.viewsets import NamedModelViewSet


class ReservedResourcesFilter(Filter):
    """
    Enables a user to filter tasks by a reserved resource href
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Queryset to filter
            value (string): href containing a reference to a reserved resource

        Returns:
            django.db.models.query.QuerySet: Queryset filtered by the reserved resource
        """

        if value is None:
            # a value was not supplied by a user
            return qs

        try:
            resolve(urlparse(value).path)
        except Resolver404:
            raise serializers.ValidationError(detail=_("URI not valid: {u}").format(u=value))

        return qs.filter(reserved_resources_record__contains=[value])


class CreatedResourcesFilter(Filter):
    """
    Filter used to get tasks by created resources.

    Created resources contain a reference to newly created repository
    versions, distributions, etc.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The QuerySet to filter
            value (string): The content href to filter by

        Returns:
            Queryset of the content contained within the specified created resource
        """

        if value is None:
            return qs

        match = resolve(value)
        resource = NamedModelViewSet.get_resource(value, match.func.cls.queryset.model)

        return qs.filter(created_resources__object_id=resource.pk)


class HyperlinkRelatedFilter(Filter):
    """
    Enables a user to filter by a foreign key using that FK's href
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("Foreign Key referenced by HREF"))
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Queryset to filter
            value (string): href containing pk for the foreign key instance

        Returns:
            django.db.models.query.QuerySet: Queryset filtered by the foreign key pk
        """

        if value is None:
            # value was not supplied by the user
            return qs

        if not value:
            raise serializers.ValidationError(
                detail=_("No value supplied for {name} filter.").format(name=self.field_name)
            )
        try:
            match = resolve(urlparse(value).path)
        except Resolver404:
            raise serializers.ValidationError(detail=_("URI not valid: {u}").format(u=value))

        pk = match.kwargs["pk"]
        try:
            UUID(pk, version=4)
        except ValueError:
            raise serializers.ValidationError(detail=_("UUID invalid: {u}").format(u=pk))

        key = "{}__pk".format(self.field_name)
        return qs.filter(**{key: pk})


class IsoDateTimeFilter(DateTimeFilter):
    """
    Uses IsoDateTimeField to support filtering on ISO 8601 formated datetimes.
    For context see:
    * https://code.djangoproject.com/ticket/23448
    * https://github.com/tomchristie/django-rest-framework/issues/1338
    * https://github.com/alex/django-filter/pull/264
    """

    field_class = IsoDateTimeField

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("ISO 8601 formatted dates are supported"))
        super().__init__(*args, **kwargs)


class RepoVersionHrefFilter(Filter):
    """
    Filter Content by a Repository Version.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("Repository Version referenced by HREF"))
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_repository_version(value):
        """
        Get the repository version from the HREF value provided by the user.

        Args:
            value (string): The RepositoryVersion href to filter by
        """
        if not value:
            raise serializers.ValidationError(
                detail=_("No value supplied for repository version filter")
            )

        return NamedModelViewSet.get_resource(value, RepositoryVersion)

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Content Queryset
            value (string): The RepositoryVersion href to filter by
        """
        raise NotImplementedError()


class RepositoryVersionFilter(RepoVersionHrefFilter):
    """
    Filter by RepositoryVersion href.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Queryset to filter
            value (string): The RepositoryVersion href to filter by

        Returns:
            Queryset filtered by given repository version on field_name
        """
        if value is None:
            # user didn't supply a value
            return qs

        repo_version = self.get_repository_version(value)
        key = f"{self.field_name}__pk"
        return qs.filter(**{key: repo_version.pk})


class ArtifactRepositoryVersionFilter(RepoVersionHrefFilter):
    """
    Filter used to get the artifacts in a repository version.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Artifact Queryset
            value (string): The RepositoryVersion href to filter by

        Returns:
            Queryset of the artifacts contained within the specified repository version
        """
        if value is None:
            # user didn't supply a value
            return qs

        repo_version = self.get_repository_version(value)
        content = repo_version.content
        artifact_pks = ContentArtifact.objects.filter(content__in=content).values("artifact__pk")
        return qs.filter(pk__in=artifact_pks)


class ContentRepositoryVersionFilter(RepoVersionHrefFilter):
    """
    Filter used to get the content of this type found in a repository version.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Content Queryset
            value (string): The RepositoryVersion href to filter by

        Returns:
            Queryset of the content contained within the specified repository version
        """
        if value is None:
            # user didn't supply a value
            return qs

        repo_version = self.get_repository_version(value)
        return qs.filter(pk__in=repo_version.content)


class ContentAddedRepositoryVersionFilter(RepoVersionHrefFilter):
    """
    Filter used to get the content of this type found in a repository version.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Content Queryset
            value (string): The RepositoryVersion href to filter by

        Returns:
            Queryset of the content added by the specified repository version
        """
        if value is None:
            # user didn't supply a value
            return qs

        repo_version = self.get_repository_version(value)
        return qs.filter(pk__in=repo_version.added())


class ContentRemovedRepositoryVersionFilter(RepoVersionHrefFilter):
    """
    Filter used to get the content of this type found in a repository version.
    """

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Content Queryset
            value (string): The RepositoryVersion href to filter by

        Returns:
            Queryset of the content removed by the specified repository version
        """
        if value is None:
            # user didn't supply a value
            return qs

        repo_version = self.get_repository_version(value)
        return qs.filter(pk__in=repo_version.removed())


class CharInFilter(BaseInFilter, CharFilter):
    pass


class LabelSelectFilter(Filter):
    """Filter to get resources that match a label filter string."""

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("help_text", _("Filter labels by search string"))
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        """
        Args:
            qs (django.db.models.query.QuerySet): The Model queryset
            value (string): label search querry

        Returns:
            Queryset of the Models filtered by label(s)

        Raises:
            rest_framework.exceptions.ValidationError: on invalid search string
        """
        if value is None:
            # user didn't supply a value
            return qs

        for term in value.split(","):
            match = re.match(r"(!?[\w\s]+)(=|!=|~)?(.*)?", term)
            if not match:
                raise DRFValidationError(_("Invalid search term: '{}'.").format(term))
            key, op, val = match.groups()

            if key.startswith("!") and op:
                raise DRFValidationError(_("Cannot use an operator with '{}'.").format(key))

            if op == "=":
                labels = Label.objects.filter(key=key, value=val)
                qs = qs.filter(pulp_labels__in=labels)
            elif op == "!=":
                labels = Label.objects.filter(key=key).exclude(value=val)
                qs = qs.filter(pulp_labels__in=labels)
            elif op == "~":
                labels = Label.objects.filter(key=key, value__icontains=val)
                qs = qs.filter(pulp_labels__in=labels)
            else:
                # 'foo', '!foo'
                if key.startswith("!"):
                    labels = Label.objects.filter(key=key[1:])
                    qs = qs.exclude(pulp_labels__in=labels)
                else:
                    labels = Label.objects.filter(key=key)
                    qs = qs.filter(pulp_labels__in=labels)

        return qs



from unittest import TestCase

from pulpcore.app.files import validate_file_paths


class TestValidateFilePaths(TestCase):
    def test_valid_paths(self):
        """
        Test for valid paths.
        """
        paths = ["a/b", "a/c/b", "PULP_MANIFEST", "b"]
        validate_file_paths(paths)

        paths = ["a/b/c", "a/b/d"]
        validate_file_paths(paths)

    def test_dupes(self):
        """
        Test for two duplicate paths.
        """
        paths = ["a/b", "PULP_MANIFEST", "PULP_MANIFEST"]
        with self.assertRaisesRegex(
            ValueError, "Path errors found. Paths are duplicated: PULP_MANIFEST"
        ):
            validate_file_paths(paths)

    def test_overlaps(self):
        """
        Test for overlapping paths.
        """
        paths = ["a/b", "a/b/c"]
        with self.assertRaises(ValueError):
            validate_file_paths(paths)

        paths = ["a/b/c", "a/b"]
        with self.assertRaises(ValueError):
            validate_file_paths(paths)

        paths = ["b/c", "a/b", "b"]
        with self.assertRaises(ValueError):
            validate_file_paths(paths)

        paths = ["a/b/c/d", "a/b"]
        with self.assertRaises(ValueError):
            validate_file_paths(paths)

        paths = ["a/b", "a/b/c/d"]
        with self.assertRaises(ValueError):
            validate_file_paths(paths)



import warnings
from gettext import gettext as _
from urllib.parse import urlparse

from django.core.exceptions import FieldDoesNotExist, FieldError, ValidationError
from django.forms.utils import ErrorList
from django.urls import Resolver404, resolve
from django_filters.rest_framework import DjangoFilterBackend, filterset
from drf_spectacular.utils import extend_schema
from guardian.shortcuts import get_objects_for_user
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.generics import get_object_or_404
from pulpcore.openapi import PulpAutoSchema
from rest_framework.serializers import ValidationError as DRFValidationError

from pulpcore.app import tasks
from pulpcore.app.models import MasterModel
from pulpcore.app.response import OperationPostponedResponse
from pulpcore.app.serializers import AsyncOperationResponseSerializer
from pulpcore.tasking.tasks import dispatch

# These should be used to prevent duplication and keep things consistent
NAME_FILTER_OPTIONS = ["exact", "in", "icontains", "contains", "startswith"]
# e.g.
# /?name=foo
# /?name__in=foo,bar
DATETIME_FILTER_OPTIONS = ["lt", "lte", "gt", "gte", "range"]
# e.g.
# /?pulp_created__gte=2018-04-12T19:45:52
# /?pulp_created__range=2018-04-12T19:45:52,2018-04-13T19:45:52


class DefaultSchema(PulpAutoSchema):
    """
    Overrides _allows_filters method to include filter fields only for read actions.

    Schema can be customised per view(set). Override this class and set it as a ``schema``
    attribute of a view(set) of interest.
    """

    def _allows_filters(self):
        """
        Include filter fields only for read actions, or GET requests.

        Returns:
            bool: True if filter fields should be included into the schema, False otherwise.
        """
        if getattr(self.view, "filter_backends", None) is None:
            return False

        if hasattr(self.view, "action"):
            return self.view.action in ["list"]

        return self.method.lower() in ["get"]


class StableOrderingFilter(OrderingFilter):
    """
    Ordering filter backend.

    Reference: https://github.com/encode/django-rest-framework/issues/6886#issuecomment-547120480
    """

    def get_ordering(self, request, queryset, view):
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting the `ordering_param` value on
        the OrderingFilter or by specifying an `ORDERING_PARAM` value in the API settings.
        """
        ordering = super(StableOrderingFilter, self).get_ordering(request, queryset, view)
        try:
            field = queryset.model._meta.get_field("pulp_created")
        except FieldDoesNotExist:
            field = queryset.model._meta.pk

        if ordering is None:
            return ["-" + field.name]

        return list(ordering) + ["-" + field.name]


class NamedModelViewSet(viewsets.GenericViewSet):
    """
    A customized named ModelViewSet that knows how to register itself with the Pulp API router.

    This viewset is discoverable by its name.
    "Normal" Django Models and Master/Detail models are supported by the ``register_with`` method.

    Attributes:
        lookup_field (str): The name of the field by which an object should be looked up, in
            addition to any parent lookups if this ViewSet is nested. Defaults to 'pk'
        endpoint_name (str): The name of the final path segment that should identify the ViewSet's
            collection endpoint.
        nest_prefix (str): Optional prefix under which this ViewSet should be nested. This must
            correspond to the "parent_prefix" of a router with rest_framework_nested.NestedMixin.
            None indicates this ViewSet should not be nested.
        parent_lookup_kwargs (dict): Optional mapping of key names that would appear in self.kwargs
            to django model filter expressions that can be used with the corresponding value from
            self.kwargs, used only by a nested ViewSet to filter based on the parent object's
            identity.
        schema (DefaultSchema): The schema class to use by default in a viewset.
    """

    endpoint_name = None
    nest_prefix = None
    parent_viewset = None
    parent_lookup_kwargs = {}
    schema = DefaultSchema()
    filter_backends = (StableOrderingFilter, DjangoFilterBackend)

    def get_serializer_class(self):
        """
        Fetch the serializer class to use for the request.

        The default behavior is to use the "serializer_class" attribute on the viewset.
        We override that for the case where a "minimal_serializer_class" attribute is defined
        and where the request contains a query parameter of "minimal=True".

        The intention is that ViewSets can define a second, more minimal serializer with only
        the most important fields.
        """
        assert self.serializer_class is not None, _(
            "'{}' should either include a `serializer_class` attribute, or override the "
            "`get_serializer_class()` method.".format(self.__class__.__name__)
        )
        minimal_serializer_class = getattr(self, "minimal_serializer_class", None)

        if minimal_serializer_class:
            if getattr(self, "request", None):
                if "minimal" in self.request.query_params:
                    # the query param is a string, and non-empty strings evaluate True,
                    # so we need to do an actual string comparison to 'true'
                    if self.request.query_params["minimal"].lower() == "true":
                        return minimal_serializer_class

        return self.serializer_class

    @staticmethod
    def get_resource(uri, model=None):
        """
        Resolve a resource URI to an instance of the resource.

        Provides a means to resolve an href passed in a POST body to an
        instance of the resource.

        Args:
            uri (str): A resource URI.
            model (django.models.Model): A model class. If not provided, the method automatically
                determines the used model from the resource URI.

        Returns:
            django.models.Model: The resource fetched from the DB.

        Raises:
            rest_framework.exceptions.ValidationError: on invalid URI or resource not found.
        """
        try:
            match = resolve(urlparse(uri).path)
        except Resolver404:
            raise DRFValidationError(detail=_("URI not valid: {u}").format(u=uri))
        if "pk" in match.kwargs:
            kwargs = {"pk": match.kwargs["pk"]}
        else:
            kwargs = {}
            for key, value in match.kwargs.items():
                if key.endswith("_pk"):
                    kwargs["{}__pk".format(key[:-3])] = value
                else:
                    kwargs[key] = value

        if model is None:
            model = match.func.cls.queryset.model

        try:
            return model.objects.get(**kwargs)
        except model.MultipleObjectsReturned:
            raise DRFValidationError(
                detail=_("URI {u} matches more than one {m}.").format(
                    u=uri, m=model._meta.model_name
                )
            )
        except model.DoesNotExist:
            raise DRFValidationError(
                detail=_("URI {u} not found for {m}.").format(u=uri, m=model._meta.model_name)
            )
        except ValidationError:
            raise DRFValidationError(detail=_("ID invalid: {u}").format(u=kwargs["pk"]))
        except FieldError:
            raise DRFValidationError(
                detail=_("URI {u} is not a valid {m}.").format(u=uri, m=model._meta.model_name)
            )

    @staticmethod
    def extract_pk(uri):
        """
        Resolve a resource URI to a simple PK value.

        Provides a means to resolve an href passed in a POST body to a primary key.
        Doesn't assume anything about whether the resource corresponding to the URI
        passed in actually exists.

        Note:
            Cannot be used with nested URIs where the PK of the final resource is not present.
            e.g. RepositoryVersion URI consists of repository PK and version number - no
            RepositoryVersion PK is present within the URI.

        Args:
            uri (str): A resource URI.

        Returns:
            primary_key (uuid.uuid4): The primary key of the resource extracted from the URI.

        Raises:
            rest_framework.exceptions.ValidationError: on invalid URI.
        """
        try:
            match = resolve(urlparse(uri).path)
        except Resolver404:
            raise DRFValidationError(detail=_("URI not valid: {u}").format(u=uri))

        try:
            return match.kwargs["pk"]
        except KeyError:
            raise DRFValidationError("URI does not contain an unqualified resource PK")

    @classmethod
    def is_master_viewset(cls):
        # ViewSet isn't related to a model, so it can't represent a master model
        if getattr(cls, "queryset", None) is None:
            return False

        # ViewSet is related to a MasterModel subclass that doesn't have its own related
        # master model, which makes this viewset a master viewset.
        if (
            issubclass(cls.queryset.model, MasterModel)
            and cls.queryset.model._meta.master_model is None
        ):
            return True

        return False

    @classmethod
    def view_name(cls):
        return "-".join(cls.endpoint_pieces())

    @classmethod
    def urlpattern(cls):
        return "/".join(cls.endpoint_pieces())

    @classmethod
    def endpoint_pieces(cls):
        # This is a core ViewSet, not Master/Detail. We can use the endpoint as is.
        if cls.queryset.model._meta.master_model is None:
            return [cls.endpoint_name]
        else:
            # Model is a Detail model. Go through its ancestry (via MRO) to find its
            # eldest superclass with a declared name, representing the Master ViewSet
            master_endpoint_name = None
            # first item in method resolution is the viewset we're starting with,
            # so start finding parents at the second item, index 1.
            for eldest in reversed(cls.mro()):
                try:
                    if eldest is not cls and eldest.endpoint_name is not None:
                        master_endpoint_name = eldest.endpoint_name
                        break
                except AttributeError:
                    # no endpoint_name defined, need to get more specific in the MRO
                    continue

            # if there is no master viewset or master endpoint name, just use endpoint_name
            if master_endpoint_name is None:
                return [cls.endpoint_name]

            # prepend endpoint of a plugin model with its Django app label
            app_label = cls.queryset.model._meta.app_label
            detail_endpoint_name = "{app_label}/{plugin_endpoint_name}".format(
                app_label=app_label, plugin_endpoint_name=cls.endpoint_name
            )

            pieces = [master_endpoint_name, detail_endpoint_name]

            # ensure that neither piece is None/empty and that they are not equal.
            if not all(pieces) or pieces[0] == pieces[1]:
                # unable to register; warn and return
                msg = (
                    "Unable to determine viewset inheritance path for master/detail "
                    "relationship represented by viewset {}. Does the Detail ViewSet "
                    "correctly subclass the Master ViewSet, and do both have endpoint_name "
                    "set to different values?"
                ).format(cls.__name__)
                warnings.warn(msg, RuntimeWarning)
                return []
            return pieces

    def initial(self, request, *args, **kwargs):
        """
        Runs anything that needs to occur prior to calling the method handler.

        For nested ViewSets, it checks that the parent object exists, otherwise return 404.
        For non-nested Viewsets, this does nothing.
        """
        if self.parent_lookup_kwargs:
            self.get_parent_field_and_object()
        super().initial(request, *args, **kwargs)

    def get_queryset(self):
        """
        Gets a QuerySet based on the current request.

        For nested ViewSets, this adds parent filters to the result returned by the superclass. For
        non-nested ViewSets, this returns the original QuerySet unchanged.

        Additional permissions-based filtering is provided for ViewSets that declare a
        ``queryset_filtering_required_permission`` attribute naming the permission users must have
        to view an object. This includes receiving the permission through either model-level,
        object-level, and access through either a user or group.

        Returns:
            django.db.models.query.QuerySet: The queryset returned by the superclass with additional
                filters applied that match self.parent_lookup_kwargs, to scope the results to only
                those associated with the parent object. Additionally the QuerySet is filtered by
                the permission named if the ViewSet declares a
                ``queryset_filtering_required_permission`` attribute.
        """
        qs = super().get_queryset()
        if self.parent_lookup_kwargs and self.kwargs:
            filters = {}
            for key, lookup in self.parent_lookup_kwargs.items():
                filters[lookup] = self.kwargs[key]
            qs = qs.filter(**filters)

        permission_name = getattr(self, "queryset_filtering_required_permission", None)
        if permission_name:
            qs = get_objects_for_user(self.request.user, permission_name, klass=qs)

        return qs

    @classmethod
    def _get_nest_depth(cls):
        """Return the depth that this ViewSet is nested."""
        if not cls.parent_lookup_kwargs:
            return 1
        return max([len(v.split("__")) for k, v in cls.parent_lookup_kwargs.items()])

    def get_parent_field_and_object(self):
        """
        For nested ViewSets, retrieve the nested parent implied by the url.

        Returns:
            tuple: (parent field name, parent)
        Raises:
            django.http.Http404: When the parent implied by the url does not exist. Synchronous
                                 use should allow this to bubble up and return a 404.
        """
        parent_field = None
        filters = {}
        if self.parent_lookup_kwargs:
            # Use the parent_lookup_kwargs and the url kwargs (self.kwargs) to retrieve the object
            for key, lookup in self.parent_lookup_kwargs.items():
                parent_field, _, parent_lookup = lookup.partition("__")
                filters[parent_lookup] = self.kwargs[key]
            return parent_field, get_object_or_404(self.parent_viewset.queryset, **filters)

    def get_parent_object(self):
        """
        For nested ViewSets, retrieve the nested parent implied by the url.

        Returns:
            pulpcore.app.models.Model: parent model object
        Raises:
            django.http.Http404: When the parent implied by the url does not exist. Synchronous
                                 use should allow this to bubble up and return a 404.
        """
        return self.get_parent_field_and_object()[1]


class AsyncReservedObjectMixin:
    """
    Mixin class providing the default method to compute the resources to reserve in the task.

    By default, lock the object instance we are working on.
    """

    def async_reserved_resources(self, instance):
        """
        Return the resources to reserve for the task created by the Async...Mixins.

        This default implementation locks the instance being worked on.

        .. note::

          This does not work for :class:`~pulpcore.app.viewsets.AsyncCreateMixin`
          (as there is no instance). Classes using :class:`~pulpcore.app.viewsets.AsyncCreateMixin`
          must override this method.

        Args:
            instance (django.models.Model): The instance that will be worked
                on by the task.

        Returns:
            list/str: The resources to put in the task's reservation

        Raises:
            AssertionError if instance is None (which happens for creation)

        """
        assert instance is not None, _(
            "'{}' must not use the default `async_reserved_resources` method "
            "when using create.".format(self.__class__.__name__)
        )
        return [instance]


class AsyncCreateMixin:
    """
    Provides a create method that dispatches a task with reservation.
    """

    @extend_schema(
        description="Trigger an asynchronous create task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def create(self, request, *args, **kwargs):
        """
        Dispatches a task with reservation for creating an instance.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        app_label = self.queryset.model._meta.app_label
        task = dispatch(
            tasks.base.general_create,
            exclusive_resources=self.async_reserved_resources(None),
            args=(app_label, serializer.__class__.__name__),
            kwargs={"data": request.data},
        )
        return OperationPostponedResponse(task, request)


class AsyncUpdateMixin(AsyncReservedObjectMixin):
    """
    Provides an update method that dispatches a task with reservation
    """

    @extend_schema(
        description="Trigger an asynchronous update task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def update(self, request, pk, **kwargs):
        partial = kwargs.pop("partial", False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        app_label = instance._meta.app_label
        task = dispatch(
            tasks.base.general_update,
            exclusive_resources=self.async_reserved_resources(instance),
            args=(pk, app_label, serializer.__class__.__name__),
            kwargs={"data": request.data, "partial": partial},
        )
        return OperationPostponedResponse(task, request)

    @extend_schema(
        description="Trigger an asynchronous partial update task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def partial_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.update(request, *args, **kwargs)


class AsyncRemoveMixin(AsyncReservedObjectMixin):
    """
    Provides a delete method that dispatches a task with reservation
    """

    @extend_schema(
        description="Trigger an asynchronous delete task",
        responses={202: AsyncOperationResponseSerializer},
    )
    def destroy(self, request, pk, **kwargs):
        """
        Delete a model instance
        """
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        app_label = instance._meta.app_label
        task = dispatch(
            tasks.base.general_delete,
            exclusive_resources=self.async_reserved_resources(instance),
            args=(pk, app_label, serializer.__class__.__name__),
        )
        return OperationPostponedResponse(task, request)


class BaseFilterSet(filterset.FilterSet):
    """
    Class to override django_filter's FilterSet and provide a way to set help text

    By default, this class will use predefined text and the field name to create help text for the
    filter. However, this can be overriden by setting a help_text dict with the the field name
    mapped to some help text:

        help_text = {'name__in': 'Lorem ipsum dolor', 'pulp_last_updated__lt': 'blah blah'}

    """

    help_text = {}

    # copied and modified from django_filter.conf
    LOOKUP_EXPR_TEXT = {
        "exact": _("matches"),
        "iexact": _("matches"),
        "contains": _("contains"),
        "icontains": _("contains"),
        "in": _("is in a comma-separated list of"),
        "gt": _("is greater than"),
        "gte": _("is greater than or equal to"),
        "lt": _("is less than"),
        "lte": _("is less than or equal to"),
        "startswith": _("starts with"),
        "istartswith": _("starts with"),
        "endswith": _("ends with"),
        "iendswith": _("ends with"),
        "range": _("is between two comma separated"),
        "isnull": _("has a null"),
        "regex": _("matches regex"),
        "iregex": _("matches regex"),
        "search": _("matches"),
        "ne": _("not equal to"),
    }

    @classmethod
    def filter_for_field(cls, field, name, lookup_expr):
        """
        Looks up and initializes a filter and returns it. Also, sets the help text on the filter.

        Args:
            field: The field class for the filter
            name: The name of filter field
            lookup_expr: The lookup expression that specifies how the field is matched
        Returns:
            django_filters.Filter: an initialized Filter object with help text
        """
        f = super().filter_for_field(field, name, lookup_expr)

        if cls.get_filter_name(name, lookup_expr) in cls.help_text:
            f.extra["help_text"] = cls.help_text[cls.get_filter_name(name, lookup_expr)]
        else:
            if lookup_expr in {"range", "in"}:
                val_word = _("values")
            else:
                val_word = _("value")

            f.extra["help_text"] = _("Filter results where {field} {expr} {value}").format(
                field=name, expr=cls.LOOKUP_EXPR_TEXT[lookup_expr], value=val_word
            )

        return f

    def is_valid(self, *args, **kwargs):
        is_valid = super().is_valid(*args, **kwargs)
        DEFAULT_FILTERS = [
            "exclude_fields",
            "fields",
            "limit",
            "minimal",
            "offset",
            "page_size",
            "ordering",
        ]
        for field in self.data.keys():
            if field in DEFAULT_FILTERS:
                continue

            if field not in self.filters:
                errors = self.form._errors.get("errors", ErrorList())
                errors.extend(["Invalid Filter: '{field}'".format(field=field)])
                self.form._errors["errors"] = errors
                is_valid = False

        return is_valid



# Generated by Django 2.2.19 on 2021-03-19 15:20

from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0061_call_handle_artifact_checksums_command'),
    ]

    operations = [
        migrations.CreateModel(
            name='Distribution',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(db_index=True, default=None)),
                ('name', models.TextField(db_index=True, unique=True)),
                ('base_path', models.TextField(unique=True)),
                ('content_guard', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ContentGuard')),
                ('publication', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Publication')),
                ('remote', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote')),
                ('repository', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Repository', related_name='distributions')),
                ('repository_version', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.RepositoryVersion')),
            ],
            options={
                'abstract': False,
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



import asyncio

import asynctest
import mock

from pulpcore.plugin.stages import Stage, EndStage, DeclarativeContent


class TestStage(asynctest.TestCase):
    def setUp(self):
        self.in_q = asyncio.Queue()
        self.stage = Stage()
        self.stage._connect(self.in_q, None)

    async def test_none_only(self):
        self.in_q.put_nowait(None)
        batch_it = self.stage.batches(minsize=1)
        with self.assertRaises(StopAsyncIteration):
            await batch_it.__anext__()

    async def test_single_batch_and_none(self):
        c1 = mock.Mock()
        c2 = mock.Mock()
        self.in_q.put_nowait(c1)
        self.in_q.put_nowait(c2)
        self.in_q.put_nowait(None)
        batch_it = self.stage.batches(minsize=1)
        self.assertEqual([c1, c2], await batch_it.__anext__())
        with self.assertRaises(StopAsyncIteration):
            await batch_it.__anext__()

    async def test_batch_and_single_none(self):
        c1 = mock.Mock()
        c2 = mock.Mock()
        self.in_q.put_nowait(c1)
        self.in_q.put_nowait(c2)
        batch_it = self.stage.batches(minsize=1)
        self.assertEqual([c1, c2], await batch_it.__anext__())
        self.in_q.put_nowait(None)
        with self.assertRaises(StopAsyncIteration):
            await batch_it.__anext__()

    async def test_two_batches(self):
        c1 = mock.Mock()
        c2 = mock.Mock()
        c3 = mock.Mock()
        c4 = mock.Mock()
        self.in_q.put_nowait(c1)
        self.in_q.put_nowait(c2)
        batch_it = self.stage.batches(minsize=1)
        self.assertEqual([c1, c2], await batch_it.__anext__())
        self.in_q.put_nowait(c3)
        self.in_q.put_nowait(c4)
        self.in_q.put_nowait(None)
        self.assertEqual([c3, c4], await batch_it.__anext__())
        with self.assertRaises(StopAsyncIteration):
            await batch_it.__anext__()

    async def test_thaw_queue(self):
        # Test that awaiting `resolved` on DeclarativeContent does not dead lock
        c1 = DeclarativeContent(mock.Mock())
        c2 = DeclarativeContent(mock.Mock())
        c3 = DeclarativeContent(mock.Mock())
        c4 = DeclarativeContent(mock.Mock())
        batch_it = self.stage.batches(minsize=4)
        fetch_task = asyncio.ensure_future(batch_it.__anext__())
        fetch_task.add_done_callback(lambda _: c2.resolve())
        self.in_q.put_nowait(c1)
        self.in_q.put_nowait(c2)
        self.in_q.put_nowait(c3)
        wait_task = asyncio.ensure_future(c2.resolution())
        wait_task.add_done_callback(lambda _: self.in_q.put_nowait(c4))
        batch, c2_res = await asyncio.gather(fetch_task, wait_task)
        self.assertEqual([c1, c2, c3], batch)
        self.assertEqual(c2.content, c2_res)
        self.in_q.put_nowait(None)
        self.assertEqual([c4], await batch_it.__anext__())
        with self.assertRaises(StopAsyncIteration):
            await batch_it.__anext__()


class TestMultipleStages(asynctest.TestCase):
    class FirstStage(Stage):
        def __init__(self, num, minsize, test_case, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.num = num
            self.minsize = minsize
            self.test_case = test_case

        async def run(self):
            for i in range(self.num):
                await asyncio.sleep(0)  # Force reschedule
                await self.put(mock.Mock())

    class MiddleStage(Stage):
        def __init__(self, num, minsize, test_case, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.num = num
            self.minsize = minsize
            self.test_case = test_case

        async def run(self):
            async for batch in self.batches(self.minsize):
                self.test_case.assertTrue(batch)
                self.test_case.assertGreaterEqual(len(batch), min(self.minsize, self.num))
                self.num -= len(batch)
                for b in batch:
                    await self.put(b)
            self.test_case.assertEqual(self.num, 0)

    class LastStage(Stage):
        def __init__(self, num, minsize, test_case, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.num = num
            self.minsize = minsize
            self.test_case = test_case

        async def run(self):
            async for batch in self.batches(self.minsize):
                self.test_case.assertTrue(batch)
                self.test_case.assertGreaterEqual(len(batch), min(self.minsize, self.num))
                self.num -= len(batch)
            self.test_case.assertEqual(self.num, 0)

    async def test_batch_queue_and_min_sizes(self):
        """Test batches iterator in a small stages setting with various sizes"""
        for num in range(10):
            for minsize in range(1, 5):
                for qsize in range(1, num + 1):
                    queues = [asyncio.Queue(maxsize=qsize) for i in range(3)]
                    first_stage = self.FirstStage(num, minsize, self)
                    middle_stage = self.MiddleStage(num, minsize, self)
                    last_stage = self.LastStage(num, minsize, self)
                    end_stage = EndStage()
                    first_stage._connect(None, queues[0])
                    middle_stage._connect(queues[0], queues[1])
                    last_stage._connect(queues[1], queues[2])
                    end_stage._connect(queues[2], None)
                    await asyncio.gather(last_stage(), middle_stage(), first_stage(), end_stage())



# Generated by Django 2.2.15 on 2020-09-02 16:16

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0044_temp_file_artifact_field'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesspolicy',
            name='permissions_assignment',
            field=django.contrib.postgres.fields.jsonb.JSONField(null=True),
        ),
    ]
